from ._http_manager import HTTPManager
from .endpoint.path import Path
from .endpoint.market import Market
from ...utils.common import Common
import json


class MarketHTTP(HTTPManager):
    async def meta(
        self,
        dex: str = None,
    ):
        """
        :param dex: str
        """
        payload = {
            "type": Market.META,
        }

        if dex is not None:
            payload["dex"] = dex

        res = await self._request(
            method="POST",
            path=Path.INFO,
            query=payload,
            signed=False,
        )
        return res

    async def spot_meta(
        self,
    ):
        payload = {
            "type": Market.SPOTMETA,
        }

        res = await self._request(
            method="POST",
            path=Path.INFO,
            query=payload,
            signed=False,
        )
        return res

    async def meta_and_asset_ctxs(
        self,
    ):
        payload = {
            "type": Market.METAANDASSETCTXS,
        }

        res = await self._request(
            method="POST",
            path=Path.INFO,
            query=payload,
            signed=False,
        )
        return res

    async def spot_meta_and_asset_ctxs(
        self,
    ):
        payload = {
            "type": Market.SPOTMETAANDASSETCTXS,
        }

        res = await self._request(
            method="POST",
            path=Path.INFO,
            query=payload,
            signed=False,
        )
        return res

    async def l2book(
        self,
        product_symbol: str,
    ):
        """
        :param product_symbol: str (e.g. BTC-USDC-SWAP)
        """
        payload = {
            "type": Market.L2BOOK,
            "coin": json.loads(self.ptm.get_exchange_symbol(product_symbol, Common.HYPERLIQUID))[0],
        }

        res = await self._request(
            method="POST",
            path=Path.INFO,
            query=payload,
            signed=False,
        )
        return res

    async def candle_snapshot(
        self,
        product_symbol: str,
        interval: str,
        startTime: int,
        endTime: int = None,
    ):
        """
        :param product_symbol: str (e.g. BTC-USDC-SWAP)
        :param interval: str (e.g. 1m, 5m, 15m, 1h, 4h, 1d)
        :param startTime: int (timestamp in milliseconds)
        :param endTime: int (timestamp in milliseconds, optional)
        """
        payload = {
            "type": Market.CANDLESNAPSHOT,
            "req": {
                "coin": json.loads(self.ptm.get_exchange_symbol(product_symbol, Common.HYPERLIQUID))[0],
                "interval": interval,
                "startTime": startTime,
            },
        }

        if endTime is not None:
            payload["req"]["endTime"] = endTime

        res = await self._request(
            method="POST",
            path=Path.INFO,
            query=payload,
            signed=False,
        )
        return res

    async def funding_rate_history(
        self,
        product_symbol: str,
        startTime: int,
        endTime: int = None,
    ):
        """
        :param product_symbol: str (e.g. BTC-USDC-SWAP)
        :param startTime: int (timestamp in milliseconds)
        :param endTime: int (timestamp in milliseconds, optional)
        """
        payload = {
            "type": Market.FUNDINGHISTORY,
            "coin": json.loads(self.ptm.get_exchange_symbol(product_symbol, Common.HYPERLIQUID))[0],
            "startTime": startTime,
        }

        if endTime is not None:
            payload["endTime"] = endTime

        res = await self._request(
            method="POST",
            path=Path.INFO,
            query=payload,
            signed=False,
        )
        return res



from enum import Enum


class Account(str, Enum):
    GET_INSTRUMENTS = "/api/v5/account/instruments"
    ACCOUNT_INFO = "/api/v5/account/balance"
    POSITION_INFO = "/api/v5/account/positions"
    POSITIONS_HISTORY = "/api/v5/account/positions-history"
    POSITION_RISK = "/api/v5/account/account-position-risk"
    BILLS_DETAIL = "/api/v5/account/bills"
    BILLS_ARCHIVE = "/api/v5/account/bills-archive"
    BILLS_HISTORY_ARCHIVE = "/api/v5/account/bills-history-archive"
    ACCOUNT_CONFIG = "/api/v5/account/config"
    POSITION_MODE = "/api/v5/account/set-position-mode"
    SET_LEVERAGE = "/api/v5/account/set-leverage"
    MAX_TRADE_SIZE = "/api/v5/account/max-size"
    MAX_AVAIL_SIZE = "/api/v5/account/max-avail-size"
    GET_LEVERAGE = "/api/v5/account/leverage-info"
    GET_ADJUST_LEVERAGE = "/api/v5/account/adjust-leverage-info"
    MAX_LOAN = "/api/v5/account/max-loan"
    FEE_RATES = "/api/v5/account/trade-fee"
    INTEREST_ACCRUED = "/api/v5/account/interest-accrued"
    INTEREST_RATE = "/api/v5/account/interest-rate"
    SET_GREEKS = "/api/v5/account/set-greeks"
    SET_ISOLATED_MODE = "/api/v5/account/set-isolated-mode"
    MAX_WITHDRAWAL = "/api/v5/account/max-withdrawal"
    INTEREST_LIMITS = "/api/v5/account/interest-limits"
    SET_AUTO_LOAN = "/api/v5/account/set-auto-loan"



import hmac
import hashlib
import time
import httpx
import logging
from dataclasses import dataclass, field
from urllib.parse import urlencode
from .endpoints.market import FuturesMarket, SpotMarket
from .endpoints.trade import FuturesTrade
from .endpoints.account import FuturesAccount
from ..product_table.manager import ProductTableManager
from ...utils.errors import FailedRequestError
from ...utils.helpers import generate_timestamp
from ...utils.common import Common


@dataclass
class HTTPManager:
    api_key: str = field(default=None)
    api_secret: str = field(default=None)
    timeout: int = field(default=10)
    logger: logging.Logger = field(default=None)
    session: httpx.AsyncClient = field(default=None, init=False)
    ptm: ProductTableManager = field(init=False)
    preload_product_table: bool = field(default=True)

    api_map = {
        "https://fapi.binance.com": {
            FuturesTrade,
            FuturesMarket,
            FuturesAccount,
        },
        "https://api.binance.com": {
            SpotMarket,
        },
    }

    async def async_init(self):
        self.session = httpx.AsyncClient(timeout=self.timeout)
        self._logger = self.logger or logging.getLogger(__name__)
        if self.preload_product_table:
            self.ptm = await ProductTableManager.get_instance(Common.BINANCE)
        return self

    def _get_base_url(self, path):
        for base_url, enums in self.api_map.items():
            if type(path) in enums:
                return base_url
        raise ValueError(f"Unknown API path: {path} (type={type(path)})")

    def _sign(self, params: dict) -> str:
        query_string = urlencode(params)
        return hmac.new(self.api_secret.encode(), query_string.encode(), hashlib.sha256).hexdigest()

    def _headers(self):
        return {"X-MBX-APIKEY": self.api_key} if self.api_key else {}

    async def _request(self, method, path, query: dict = None, signed: bool = True):
        if not self.session:
            await self.async_init()

        if query is None:
            query = {}

        if signed:
            if not (self.api_key and self.api_secret):
                raise ValueError("Signed request requires API Key and Secret.")
            query["timestamp"] = int(time.time() * 1000)
            query["recvWindow"] = 5000
            query["signature"] = self._sign(query)

        response = None
        try:
            base_url = self._get_base_url(path)
            url = f"{base_url}{path}"
            if method.upper() == "GET":
                url += f"?{urlencode(query)}" if query else ""
                response = await self.session.get(url, headers=self._headers())
            elif method.upper() == "POST":
                response = await self.session.post(url, headers=self._headers(), data=query)
            elif method.upper() == "DELETE":
                url += f"?{urlencode(query)}" if query else ""
                response = await self.session.delete(url, headers=self._headers())
            else:
                raise ValueError(f"Unsupported method: {method}")

            try:
                data = response.json()
            except Exception:
                data = {}

            timestamp = generate_timestamp(iso_format=True)

            if isinstance(data, dict) and "code" in data and str(data["code"]) != "200":
                code = data.get("code", "Unknown")
                error_message = data.get("msg", "Unknown error")
                raise FailedRequestError(
                    request=f"{method} {url} | Body: {query}",
                    message=f"BINANCE API Error: [{code}] {error_message}",
                    status_code=response.status_code,
                    time=timestamp,
                    resp_headers=response.headers,
                )

            if not response.status_code // 100 == 2:
                raise FailedRequestError(
                    request=f"{method.upper()} {url} | Body: {query}",
                    message=f"HTTP Error {response.status_code}: {response.text}",
                    status_code=response.status_code,
                    time=timestamp,
                    resp_headers=response.headers,
                )

            return data

        except httpx.RequestError as e:
            raise FailedRequestError(
                request=f"{method.upper()} {url} | Params: {query}",
                message=f"Request failed: {str(e)}",
                status_code=response.status_code if response else "Unknown",
                time=query.get("timestamp", "Unknown"),
                resp_headers=response.headers if response else None,
            )



from ._http_manager import HTTPManager
from .endpoints.account import FundingAccount, FuturesAccount


class AccountHTTP(HTTPManager):
    def get_account_balance(
        self,
        currency: str = None,
        needUsdValuation: bool = False,
    ):
        """
        :param currency: str
        :param needUsdValuation: bool
        """
        payload = {
            "needUsdValuation": needUsdValuation,
        }
        if currency is not None:
            payload["currency"] = currency

        res = self._request(
            method="GET",
            path=FundingAccount.GET_ACCOUNT_BALANCE,
            query=payload,
        )
        return res

    def get_account_currencies(
        self,
        currencies: str = None,
    ):
        """
        :param currencies: str
        """
        payload = {}
        if currencies is not None:
            coinName = ",".join(currencies)
            payload = {
                "currencies": coinName,
            }

        res = self._request(
            method="GET",
            path=FundingAccount.GET_ACCOUNT_CURRENCIES,
            query=payload,
        )
        return res

    def get_spot_wallet(self):
        res = self._request(
            method="GET",
            path=FundingAccount.GET_SPOT_WALLET_BALANCE,
            query=None,
        )
        return res

    def get_deposit_address(
        self,
        currency: str,
    ):
        """
        :param currency: str
        """
        payload = {
            "currency": currency,
        }

        res = self._request(
            method="GET",
            path=FundingAccount.DEPOSIT_ADDRESS,
            query=payload,
        )
        return res

    def get_withdraw_charge(
        self,
        currency: str,
    ):
        """
        :param currency: str
        """
        payload = {
            "currency": currency,
        }

        res = self._request(
            method="GET",
            path=FundingAccount.WITHDRAW_QUOTA,
            query=payload,
        )
        return res

    def post_withdraw_apply(
        self,
        currency: str,
        amount: str,
    ):
        """
        :param currency: str
        :param amount: str
        """
        payload = {
            "currency": currency,
            "amount": amount,
        }

        res = self._request(
            method="POST",
            path=FundingAccount.WITHDRAW,
            query=payload,
        )
        return res

    def get_deposit_withdraw_history(
        self,
        operation_type: str = "withdraw",
        limit: int = 1000,
        currency: str = None,
        startTime: int = None,
        endTime: int = None,
    ):
        """
        :param limit: int
        :param currency: str
        :param operation_type: str (deposit, withdraw)
        :param startTime: int
        :param endTime: int
        """
        payload = {
            "N": limit,
            "operation_type": operation_type,
        }
        if currency is not None:
            payload["currency"] = currency
        if startTime is not None:
            payload["startTime"] = startTime
        if endTime is not None:
            payload["endTime"] = endTime

        res = self._request(
            method="GET",
            path=FundingAccount.GET_DEPOSIT_WITHDRAW_HISTORY,
            query=payload,
        )
        return res

    def get_deposit_withdraw_history_detail(
        self,
        id: str,
    ):
        """
        :param id: str (withdraw_id or deposit_id)
        """
        payload = {
            "id": id,
        }

        res = self._request(
            method="GET",
            path=FundingAccount.GET_DEPOSIT_WITHDRAW_HISTORY_DETAIL,
            query=payload,
        )
        return res

    def get_contract_assets(self):
        res = self._request(
            method="GET",
            path=FuturesAccount.GET_CONTRACT_ASSETS,
            query=None,
        )
        return res



from ._http_manager import HTTPManager
from .endpoints.path import Path
from .endpoints.account import Account


class AccountHTTP(HTTPManager):
    def clearinghouse_state(
        self,
        user: str,
        dex: str = None,
    ):
        """
        :param user: str (waller address)
        :param dex: str
        """
        payload = {
            "type": Account.CLEARINGHOUSESTATE,
            "user": user,
        }

        if dex is not None:
            payload["dex"] = dex

        res = self._request(
            method="POST",
            path=Path.INFO,
            query=payload,
            signed=False,
        )
        return res

    def open_orders(
        self,
        user: str,
        dex: str = None,
    ):
        """
        :param user: str (waller address)
        :param dex: str
        """
        payload = {
            "type": Account.OPENORDERS,
            "user": user,
        }

        if dex is not None:
            payload["dex"] = dex

        res = self._request(
            method="POST",
            path=Path.INFO,
            query=payload,
            signed=False,
        )
        return res

    def user_fills(
        self,
        user: str,
        aggregateByTime: bool = False,
    ):
        """
        :param user: str (waller address)
        :param aggregateByTime: bool
        """
        payload = {
            "type": Account.USERFILLS,
            "user": user,
        }

        if aggregateByTime:
            payload["aggregateByTime"] = True

        res = self._request(
            method="POST",
            path=Path.INFO,
            query=payload,
            signed=False,
        )
        return res

    def user_rate_limit(
        self,
        user: str,
    ):
        """
        :param user: str (waller address)
        """
        payload = {
            "type": Account.USERRATELIMIT,
            "user": user,
        }

        res = self._request(
            method="POST",
            path=Path.INFO,
            query=payload,
            signed=False,
        )
        return res

    def order_status(
        self,
        user: str,
        oid: str,
    ):
        """
        :param user: str (waller address)
        :param oid: str
        """
        payload = {
            "type": Account.ORDERSTATUS,
            "user": user,
            "oid": oid,
        }

        res = self._request(
            method="POST",
            path=Path.INFO,
            query=payload,
            signed=False,
        )
        return res

    def historical_orders(
        self,
        user: str,
    ):
        """
        :param user: str (waller address)
        """
        payload = {
            "type": Account.HISTORICALORDERS,
            "user": user,
        }

        res = self._request(
            method="POST",
            path=Path.INFO,
            query=payload,
            signed=False,
        )
        return res

    def subaccounts(
        self,
        user: str,
    ):
        """
        :param user: str (waller address)
        """
        payload = {
            "type": Account.SUBACCOUNTS,
            "user": user,
        }

        res = self._request(
            method="POST",
            path=Path.INFO,
            query=payload,
            signed=False,
        )
        return res

    def user_role(
        self,
        user: str,
    ):
        """
        :param user: str (waller address)
        """
        payload = {
            "type": Account.USERROLE,
            "user": user,
        }

        res = self._request(
            method="POST",
            path=Path.INFO,
            query=payload,
            signed=False,
        )
        return res

    def portfolio(
        self,
        user: str,
    ):
        """
        :param user: str (waller address)
        """
        payload = {
            "type": Account.PORTFOLIO,
            "user": user,
        }

        res = self._request(
            method="POST",
            path=Path.INFO,
            query=payload,
            signed=False,
        )
        return res



from dataclasses import dataclass
from enum import Enum


class PositionSide(str, Enum):
    LONG = "long"
    SHORT = "short"
    NONE = "none"  # okx 轉入保證金後會產生pos為0的情況


@dataclass
class PositionInfo:
    symbol: str
    side: PositionSide
    quantity: float
    open_avg_price: float
    # closed_volume: float



import json
import time
import hmac
import hashlib
import logging
import polars as pl
from krex.async_support.product_table.manager import ProductTableManager
from .models.balance import BalanceInfo
from .base import WsClient
from .data.market_data import MarketData
from .models.ticker import BookTicker
from .models.order import Exchange, OrderStatus, TradeStatus

logger = logging.getLogger(__name__)


class BitmartPublicWsClient(WsClient):
    URI = "wss://openapi-ws-v2.bitmart.com/api?protocol=1.1"
    SANDBOX_URI = ""

    def __init__(
        self,
        subscription: dict,
        api_key: str,
        api_secret: str,
        memo: str,
        is_sandbox: bool = False,
        slack=None,
        slack_bot_name: str = None,
        slack_channel_name: str = None,
    ):
        super().__init__(
            subscription=subscription,
            is_sandbox=is_sandbox,
            slack=slack,
            slack_bot_name=slack_bot_name,
            slack_channel_name=slack_channel_name,
        )
        self.api_key = api_key
        self.api_secret = api_secret
        self.memo = memo

        self.market_data = MarketData()
        self.ptm = None

    @classmethod
    async def create(cls, **kwargs):
        self = cls(**kwargs)
        self.ptm = await ProductTableManager.get_instance()
        return self

    def sign(self):
        timestamp = str(int(time.time() * 1000))
        sign_payload = f"{timestamp}#{self.memo}#bitmart.WebSocket"
        signature = hmac.new(
            self.api_secret.encode("utf-8"),
            sign_payload.encode("utf-8"),
            hashlib.sha256,
        ).hexdigest()
        return {
            "action": "access",
            "args": [self.api_key, timestamp, signature, "web"],
        }

    async def on_open(self):
        await super().on_open()

        sign_message = json.dumps(self.sign())
        await self.websocket.send(sign_message)

        while True:
            message = await self.websocket.recv()
            payload = json.loads(message)
            if payload.get("action") == "access" and payload.get("success"):
                break
            logger.debug(f"Waiting for access confirmation: {payload}")

        subscribe_message = json.dumps(self.subscription)
        await self.websocket.send(subscribe_message)
        logger.info(f"Sent subscription message: {subscribe_message}")

    async def on_message(self, message: str):
        await super().on_message(message)
        payload = json.loads(message)

        if payload.get("data") and payload.get("group", "").startswith("futures/bookticker:"):
            data_payload = payload["data"]
            data = BookTicker(
                symbol=data_payload.get("symbol", ""),
                best_bid_price=float(data_payload.get("best_bid_price", 0)),
                best_bid_quantity=float(data_payload.get("best_bid_vol", 0)),
                best_ask_price=float(data_payload.get("best_ask_price", 0)),
                best_ask_quantity=float(data_payload.get("best_ask_vol", 0)),
                timestamp=int(data_payload.get("ms_t", 0)),
            )
            await self.market_data.update_depth_data("bitmart", data.symbol, data)

        elif payload.get("data") and payload.get("group") == "futures/order":
            for order_payload in payload["data"]:
                try:
                    order_data = order_payload.get("order", {})
                    state = order_data["state"]
                    size = float(order_data["size"])
                    deal_size = float(order_data["deal_size"])

                    if state == 4:
                        status = OrderStatus.CANCELLED
                    elif deal_size == 0:
                        status = OrderStatus.OPEN
                    elif deal_size == size:
                        status = OrderStatus.FILLED
                    elif size > deal_size:
                        status = OrderStatus.PARTIALLY_FILLED
                    else:
                        status = OrderStatus.OPEN

                    trade_status = TradeStatus(
                        order_id=order_data["order_id"],
                        client_oid=order_data["client_order_id"],
                        order_type=order_data["type"],
                        status=status,
                        price=float(order_data.get("price", 0.0)),
                        quantity=float(order_data.get("size", 0.0)),
                        filled_quantity=float(order_data.get("deal_size", 0.0)),
                        deal_avg_price=float(order_data.get("deal_avg_price", 0.0)),
                    )

                    await self.market_data.update_trade_status(Exchange.BITMART, order_data["symbol"], trade_status)
                    logger.info(f"Updated TradeStatus: {trade_status}")

                except Exception as e:
                    logger.error(f"Error processing order data: {e}")

        elif payload.get("data") and payload.get("group", "").startswith("futures/klineBin"):
            try:
                data_payload = payload["data"]["items"][0]
                df = pl.DataFrame(
                    {
                        "open": [float(data_payload["o"])],
                        "high": [float(data_payload["h"])],
                        "low": [float(data_payload["l"])],
                        "close": [float(data_payload["c"])],
                        "volume": [float(data_payload["v"])],
                        "datetime": [int(data_payload["ts"])],
                    }
                )
                df = df.with_columns([pl.col("datetime").cast(pl.Datetime).dt.cast_time_unit("ms")])
                df = df.set_sorted("datetime")

                product_symbol = self.ptm.get_product_symbol(payload["data"]["symbol"], "bitmart")
                await self.market_data.update_kline_data("bitmart", product_symbol, df)
            except Exception as e:
                logger.error(f"Error processing kline data: {e}")

        elif payload.get("group", "").startswith("futures/asset:"):
            data = payload.get("data")
            try:
                balance_items = data if isinstance(data, list) else [data]
                for item in balance_items:
                    balance_info = BalanceInfo(
                        symbol=item["currency"],
                        available_balance=float(item.get("available_balance", 0.0)),
                        position_deposit=float(item.get("position_deposit", 0.0)),
                        frozen_balance=float(item.get("frozen_balance", 0.0)),
                    )
                    await self.market_data.update_balance(Exchange.BITMART, item["currency"], balance_info)
                    logger.info(f"Updated BalanceInfo: {balance_info}")
            except Exception as e:
                logger.error(f"Error processing balance data: {e}, payload: {payload}")



import asyncio
import polars as pl
from contextlib import asynccontextmanager
from .fetch import bybit, okx, bitmart, gateio, binance, hyperliquid

VALID_EXCHANGES = [
    bybit,
    okx,
    bitmart,
    gateio,
    binance,
    hyperliquid,
]


class ProductTableError(Exception):
    pass


class ProductTableManager:
    """
    Exchange Product Mapping Table

    This table provides a structured mapping between product symbols and their corresponding exchange-specific symbols,
    along with key trading attributes. It helps standardize the representation of products across different exchanges.

    Columns:
        - product_symbol: Standardized product identifier used internally.
        - exchange_symbol: The product symbol as recognized on the exchange.
        - exchange: The name of the exchange where the product is traded.
        - product_type: The category of the product (e.g., SPOT, SWAP, FUTURES).
        - price_precision: The decimal precision allowed for price values (if applicable).
        - size_precision: The decimal precision allowed for order sizes (if applicable).
        - contract_value: The notional value of one contract (for derivatives).
        - min_size: The minimum order size allowed on the exchange.
        - min_notional: The minimum notional value required for an order.

    Example:
        | product_symbol  | exchange_symbol | exchange | product_type | price_precision | size_precision | contract_value | min_size | min_notional |
        |-----------------|-----------------|----------|--------------|-----------------|----------------|----------------|----------|--------------|
        | BTC-USDT-SPOT   | BTC/USDT        | BINANCE  | SPOT         | 0.01            | 0.0001         | N/A            | 0.0001   | 10           |
        | BTC-USD-SWAP    | BTCUSDT         | BINANCE  | SWAP         | 0.1             | 0.001          | 100 USD        | 0.001    | 5            |
        | BTC-USD-FUTURES | BTC-USD-SWAP    | OKX      | FUTURES      | 0.01            | 0.0001         | 10 USD         | 0.0001   | 1            |

    Use this mapping to correctly interpret product symbols and their attributes when integrating with multiple exchanges.
    """

    _instance = {}

    @classmethod
    async def get_instance(cls, exchange_name=None):
        if exchange_name not in cls._instance:
            cls._instance[exchange_name] = cls()
            await cls._instance[exchange_name]._initialize(exchange_name=exchange_name)
        return cls._instance[exchange_name]

    async def _initialize(self, exchange_name=None):
        """Initialize the product table by fetching data from valid exchanges."""
        self.product_table = await self._fetch_product_tables(exchange_name)

    async def _fetch_product_tables(self, exchange_name=None):
        """Fetch product tables from all valid exchanges and combine them into a single DataFrame."""
        if exchange_name is None:
            product_tables = await asyncio.gather(*[func() for func in VALID_EXCHANGES])
        else:
            product_tables = await asyncio.gather(*[func() for func in VALID_EXCHANGES if func.__name__ == exchange_name])
        return pl.concat(product_tables, how="vertical")

    @asynccontextmanager
    async def _create_exchange_clients(self):
        """Create exchange clients as an async context manager to ensure proper cleanup."""
        clients = [exchange() for exchange in VALID_EXCHANGES]
        try:
            yield clients
        finally:
            await asyncio.gather(*[client.close() for client in clients if hasattr(client, "close")])

    def get(
        self,
        key,
        product_symbol=None,
        exchange=None,
        product_type=None,
        exchange_symbol=None,
    ):
        """
        Return a array of values of key from product that satisfy the conditions.
        The conditions are case-insensitive (except id_).
        """
        data = self.product_table
        if product_symbol is not None:
            data = data.filter(pl.col("product_symbol") == product_symbol)
        if exchange is not None:
            data = data.filter(pl.col("exchange") == exchange)
        if product_type is not None:
            data = data.filter(pl.col("product_type") == product_type)
        if exchange_symbol is not None:
            data = data.filter(pl.col("exchange_symbol") == exchange_symbol)

        if data.height > 1:
            raise ProductTableError(
                f"Exist multiple {key} for product_symbol: {product_symbol}, exchange: {exchange}, product_type: {product_type}"
            )
        if data.height == 0:
            raise ProductTableError(
                f"Not exist {key} for product_symbol: {product_symbol}, exchange: {exchange}, product_type: {product_type}, exchange_symbol: {exchange_symbol}"
            )

        return data.select(key).item()

    def get_exchange_symbol(self, product_symbol, exchange):
        return self.get("exchange_symbol", product_symbol, exchange)

    def get_product_symbol(self, exchange_symbol, exchange, product_type=None):
        if product_type is None:
            return self.get("product_symbol", exchange_symbol=exchange_symbol, exchange=exchange)
        else:
            return self.get("product_symbol", exchange_symbol=exchange_symbol, exchange=exchange, product_type=product_type)

    def get_product_type(self, product_symbol, exchange):
        return self.get("product_type", product_symbol, exchange)

    def get_base_currency(self, product_symbol, exchange):
        return self.get("base_currency", product_symbol, exchange)

    def get_quote_currency(self, product_symbol, exchange):
        return self.get("quote_currency", product_symbol, exchange)

    def get_trading_details(self, product_symbol, exchange):
        return {
            "price_precision": self.get("price_precision", product_symbol, exchange),
            "size_precision": self.get("size_precision", product_symbol, exchange),
            "min_size": self.get("min_size", product_symbol, exchange),
            "min_notional": self.get("min_notional", product_symbol, exchange),
            "size_per_contract": self.get("size_per_contract", product_symbol, exchange),
        }

    def get_exchange_symbols(self, exchange, product_type=None):
        if product_type is None:
            return self.product_table.filter(pl.col("exchange") == exchange).select("exchange_symbol").to_series().to_list()
        else:
            return self.product_table.filter(pl.col("exchange") == exchange).filter(pl.col("product_type") == product_type).select("exchange_symbol").to_series().to_list()
    
    def get_product_symbols(self, exchange, product_type=None):
        if product_type is None:
            return self.product_table.filter(pl.col("exchange") == exchange).select("product_symbol").to_series().to_list()
        else:
            return self.product_table.filter(pl.col("exchange") == exchange).filter(pl.col("product_type") == product_type).select("product_symbol").to_series().to_list()


import json
import time
import hmac
import hashlib
import base64
import logging
import polars as pl
from .models.balance import BalanceInfo
from .models.position import PositionInfo, PositionSide
from .base import WsClient
from .data.market_data import MarketData
from .models.ticker import BookTicker
from .models.order import Exchange, OrderStatus, TradeStatus

logger = logging.getLogger(__name__)


def safe_float(value, default=0.0):
    if value is None or value == "":
        return default
    try:
        return float(value)
    except (ValueError, TypeError):
        return default


class OkxPublicWsClient(WsClient):
    URI = "wss://ws.okx.com:8443/ws/v5/public"

    def __init__(
        self,
        subscription: dict,
        is_sandbox: bool = False,
        orderbook_queue=None,
        slack=None,
        slack_bot_name: str = None,
        slack_channel_name: str = None,
    ):
        super().__init__(subscription, is_sandbox, orderbook_queue, slack, slack_bot_name, slack_channel_name)
        self.market_data = MarketData()

    @classmethod
    async def create(cls, **kwargs):
        self = cls(**kwargs)
        return self

    async def on_message(self, message: str):
        await super().on_message(message)
        payload = json.loads(message)

        channel = payload.get("arg", {}).get("channel")
        data = payload.get("data", [])

        if channel == "bbo-tbt" and data:
            data_payload = data[0]
            book_ticker = BookTicker(
                symbol=payload["arg"].get("instId"),
                product_type="SWAP" if payload["arg"].get("instId").endswith("-SWAP") else "SPOT",
                best_bid_price=float(data_payload.get("bids", [[0, 0]])[0][0]),
                best_bid_quantity=float(data_payload.get("bids", [[0, 0]])[0][1]),
                best_ask_price=float(data_payload.get("asks", [[0, 0]])[0][0]),
                best_ask_quantity=float(data_payload.get("asks", [[0, 0]])[0][1]),
                timestamp=int(data_payload.get("ts", 0)),
            )
            await self.market_data.update_depth_data("okx", book_ticker.symbol, book_ticker)
            await self.update_orderbook_queue("okx", book_ticker.symbol, book_ticker)

        elif channel and "candle" in channel and data:
            kline_payload = data[0]
            symbol = payload["arg"].get("instId", "")

            new_df = (
                pl.DataFrame(
                    {
                        "open": [float(kline_payload[1])],
                        "high": [float(kline_payload[2])],
                        "low": [float(kline_payload[3])],
                        "close": [float(kline_payload[4])],
                        "volume": [float(kline_payload[5])],
                        "datetime": [int(kline_payload[0])],
                    }
                )
                .with_columns(pl.col("datetime").cast(pl.Int64).cast(pl.Datetime("ms")))
                .set_sorted("datetime")
            )

            # 直接送到 update_kline_data，由 shared_data 去處理要怎麼存
            await self.market_data.update_kline_data("okx", symbol, new_df)


class OkxPrivateWsClient(WsClient):
    URI = "wss://ws.okx.com:8443/ws/v5/private"

    def __init__(
        self,
        subscription: dict,
        api_key: str,
        api_secret: str,
        passphrase: str,
        is_sandbox: bool = False,
        slack=None,
        slack_bot_name: str = None,
        slack_channel_name: str = None,
    ):
        super().__init__(subscription, is_sandbox, slack, slack_bot_name, slack_channel_name)
        self.api_key = api_key
        self.api_secret = api_secret
        self.passphrase = passphrase
        self.market_data = MarketData()

    @classmethod
    async def create(cls, **kwargs):
        self = cls(**kwargs)
        return self

    def sign(self):
        timestamp = str(time.time())
        message = timestamp + "GET" + "/users/self/verify"
        mac = hmac.new(self.api_secret.encode(), message.encode(), hashlib.sha256)
        sign = base64.b64encode(mac.digest()).decode()
        return {
            "op": "login",
            "args": [
                {
                    "apiKey": self.api_key,
                    "passphrase": self.passphrase,
                    "timestamp": timestamp,
                    "sign": sign,
                }
            ],
        }

    async def on_open(self):
        await super().on_open()

        sign_message = json.dumps(self.sign())
        await self.websocket.send(sign_message)

        while True:
            message = await self.websocket.recv()
            payload = json.loads(message)
            if payload.get("event") == "login" and payload.get("code") == "0":
                break
            logger.debug(f"Waiting for login confirmation: {payload}")

        subscribe_message = json.dumps(self.subscription)
        await self.websocket.send(subscribe_message)
        logger.info(f"Sent subscription message: {subscribe_message}")

    async def on_message(self, message: str):
        await super().on_message(message)
        payload = json.loads(message)

        channel = payload.get("arg", {}).get("channel")
        data = payload.get("data", [])

        if channel == "orders" and data:
            for order_data in data:
                try:
                    status_mapping = {
                        "live": OrderStatus.OPEN,
                        "partially_filled": OrderStatus.PARTIALLY_FILLED,
                        "filled": OrderStatus.FILLED,
                        "canceled": OrderStatus.CANCELLED,
                    }
                    trade_status = TradeStatus(
                        order_id=order_data["ordId"],
                        client_oid=order_data["clOrdId"],
                        order_type=order_data["ordType"],
                        status=status_mapping.get(order_data["state"], OrderStatus.OPEN),
                        price=float(order_data.get("px", 0.0)),
                        quantity=float(order_data.get("sz", 0.0)),
                        filled_quantity=float(order_data.get("fillSz", 0.0)),
                        deal_avg_price=float(order_data.get("avgPx", 0.0)),
                    )
                    await self.market_data.update_trade_status(Exchange.OKX, order_data["instId"], trade_status)
                    logger.info(f"Updated TradeStatus: {trade_status}")
                except Exception as e:
                    logger.error(f"Error processing order data: {e}")

        elif channel == "positions" and data:
            for position_data in data:
                try:
                    side_mapping = {
                        "long": PositionSide.LONG,
                        "short": PositionSide.SHORT,
                        "net": PositionSide.NONE,
                    }
                    side = side_mapping.get(position_data["posSide"], PositionSide.NONE)
                    pos_value = float(position_data.get("pos", 0.0))

                    if side == PositionSide.NONE:
                        side = (
                            PositionSide.LONG
                            if pos_value > 0
                            else (PositionSide.SHORT if pos_value < 0 else PositionSide.NONE)
                        )

                    position_info = PositionInfo(
                        symbol=position_data["instId"],
                        side=side,
                        quantity=abs(pos_value),
                        open_avg_price=float(position_data.get("avgPx", 0.0)),
                    )
                    await self.market_data.update_position(Exchange.OKX, position_data["instId"], position_info)
                    logger.info(f"Updated PositionInfo: {position_info}")
                except Exception as e:
                    logger.error(f"Error processing position data: {e}")

        elif channel == "account" and data:
            for balance_data in data:
                try:
                    details = balance_data.get("details", [])
                    for detail in details:
                        balance_info = BalanceInfo(
                            symbol=detail["ccy"],
                            available_balance=safe_float(detail.get("availBal")),
                            position_deposit=safe_float(detail.get("imr")),
                            frozen_balance=safe_float(detail.get("frozenBal")),
                        )
                        await self.market_data.update_balance(Exchange.OKX, detail["ccy"], balance_info)
                        logger.info(f"Updated BalanceInfo: {balance_info}")
                except Exception as e:
                    logger.error(f"Error processing balance data: {e}")



from dataclasses import dataclass


@dataclass
class BookTicker:
    symbol: str
    product_type: str
    best_bid_price: float
    best_bid_quantity: float
    best_ask_price: float
    best_ask_quantity: float
    timestamp: int



from enum import Enum


class Market(str, Enum):
    META = "meta"
    SPOTMETA = "spotMeta"
    METAANDASSETCTXS = "metaAndAssetCtxs"
    SPOTMETAANDASSETCTXS = "spotMetaAndAssetCtxs"
    L2BOOK = "l2Book"
    CANDLESNAPSHOT = "candleSnapshot"
    FUNDINGHISTORY = "fundingHistory"

    def __str__(self) -> str:
        return self.value

    def __repr__(self) -> str:
        return repr(self.value)



from enum import Enum


class Account(str, Enum):
    CLEARINGHOUSESTATE = "clearinghouseState"
    OPENORDERS = "openOrders"
    USERFILLS = "userFills"
    USERRATELIMIT = "userRateLimit"
    ORDERSTATUS = "orderStatus"
    HISTORICALORDERS = "historicalOrders"
    SUBACCOUNTS = "subaccounts"
    USERROLE = "userRole"
    PORTFOLIO = "portfolio"

    def __str__(self) -> str:
        return self.value

    def __repr__(self) -> str:
        return repr(self.value)



import uuid
import time
from ._http_manager import HTTPManager
from .endpoints.asset import Asset


class AssetHTTP(HTTPManager):
    async def get_coin_info(
        self,
        coin: str = None,
    ):
        """
        :param coin: str
        """
        payload = {}
        if coin is not None:
            payload["coin"] = coin

        res = await self._request(
            method="GET",
            path=Asset.GET_COIN_INFO,
            query=payload,
        )
        return res

    async def get_sub_uid(self):
        res = await self._request(
            method="GET",
            path=Asset.GET_SUB_UID,
            query=None,
        )
        return res

    async def get_spot_asset_info(
        self,
        # accountType: str,
        coin: str = None,
    ):
        """
        async default accountType: SPOT
        :param coin: str
        """
        payload = {
            "accountType": "SPOT",
        }
        if coin is not None:
            payload["coin"] = coin

        res = await self._request(
            method="GET",
            path=Asset.GET_SPOT_ASSET_INFO,
            query=payload,
        )
        return res

    async def get_coins_balance(
        self,
        accountType: str,
        coin: str = None,
        memberId: str = None,
    ):
        """
        :param accountType: str
        :param coin: str
        :param memberId: str
        """
        payload = {
            "accountType": accountType,
        }
        if coin is not None:
            payload["coin"] = coin
        if memberId is not None:
            payload["memberId"] = memberId

        res = await self._request(
            method="GET",
            path=Asset.GET_ALL_COINS_BALANCE,
            query=payload,
        )
        return res

    async def get_coin_balance(
        self,
        accountType: str,
        coin: str,
        memberId: str = None,
        toAccountType: str = None,
    ):
        """
        :param accountType: str
        :param coin: str
        :param memberId: str
        :param toAccountType: str
        """
        payload = {
            "accountType": accountType,
            "coin": coin,
        }
        if memberId is not None:
            payload["memberId"] = memberId
        if toAccountType is not None:
            payload["toAccountType"] = toAccountType

        res = await self._request(
            method="GET",
            path=Asset.GET_SINGLE_COIN_BALANCE,
            query=payload,
        )
        return res

    async def get_withdrawable_amount(
        self,
        coin: str,
    ):
        """
        :param coin: str
        """
        payload = {
            "coin": coin,
        }

        res = await self._request(
            method="GET",
            path=Asset.GET_WITHDRAWABLE_AMOUNT,
            query=payload,
        )
        return res

    async def get_internal_transfer_records(
        self,
        coin: str = None,
        startTime: int = None,
        limit: int = 20,
    ):
        """
        :param coin: str
        :param startTime: int
        :param limit: int
        """
        payload = {
            "limit": limit,
        }
        if coin is not None:
            payload["coin"] = coin
        if startTime is not None:
            payload["startTime"] = startTime

        res = await self._request(
            method="GET",
            path=Asset.GET_INTERNAL_TRANSFER_RECORDS,
            query=payload,
        )
        return res

    async def get_transferable_coin(
        self,
        fromAccountType: str,
        toAccountType: str,
    ):
        """ "
        :param fromAccountType: str
        :param toAccountType: str
        """
        payload = {
            "fromAccountType": fromAccountType,
            "toAccountType": toAccountType,
        }

        res = await self._request(
            method="GET",
            path=Asset.GET_TRANSFERABLE_COIN,
            query=payload,
        )
        return res

    async def create_internal_transfer(
        self,
        coin: str,
        amount: str,
        fromAccountType: str,
        toAccountType: str,
    ):
        """
        :param coin: str
        :param amount: str
        :param fromAccountType: str
        :param toAccountType: str
        """
        transfer_id = str(uuid.uuid4())
        payload = {
            "transferId": transfer_id,
            "coin": coin,
            "amount": amount,
            "fromAccountType": fromAccountType,
            "toAccountType": toAccountType,
        }

        res = await self._request(
            method="POST",
            path=Asset.CREATE_INTERNAL_TRANSFER,
            query=payload,
        )
        return res

    async def create_universal_transfer(
        self,
        coin: str,
        amount: str,
        fromMemberId: int,
        toMemberId: int,
        fromAccountType: str,
        toAccountType: str,
    ):
        """
        :coin: str
        :amount: str
        :fromMemberId: int
        :toMemberId: int
        :fromAccountType: str
        :toAccountType: str
        """
        transfer_id = str(uuid.uuid4())
        payload = {
            "transferId": transfer_id,
            "coin": coin,
            "amount": amount,
            "fromMemberId": fromMemberId,
            "toMemberId": toMemberId,
            "fromAccountType": fromAccountType,
            "toAccountType": toAccountType,
        }

        res = await self._request(
            method="POST",
            path=Asset.CREATE_UNIVERSAL_TRANSFER,
            query=payload,
        )
        return res

    async def get_universal_transfer_records(
        self,
        coin: str = None,
        status: str = None,
        startTime: int = None,
        limit: int = 20,
    ):
        """
        :param coin: str
        :param status: str
        :param startTime: int
        :param limit: int
        """
        payload = {
            "limit": limit,
        }
        if coin is not None:
            payload["coin"] = coin
        if status is not None:
            payload["status"] = status
        if startTime is not None:
            payload["startTime"] = startTime

        res = await self._request(
            method="GET",
            path=Asset.GET_UNIVERSAL_TRANSFER_RECORDS,
            query=payload,
        )
        return res

    async def set_deposit_account(
        self,
        accountType: str,
    ):
        """
        :param accountType: str
        :param coin: str
        """
        payload = {
            "accountType": accountType,
        }

        res = await self._request(
            method="POST",
            path=Asset.SET_DEPOSIT_ACCOUNT,
            query=payload,
        )
        return res

    async def get_deposit_records(
        self,
        coin: str = None,
        startTime: int = None,
        limit: int = 20,
    ):
        """
        :param coin: str
        :param startTime: str
        :param limit: int
        """
        payload = {
            "limit": limit,
        }
        if coin is not None:
            payload["coin"] = coin
        if startTime is not None:
            payload["startTime"] = startTime

        res = await self._request(
            method="GET",
            path=Asset.GET_DEPOSIT_RECORDS,
            query=payload,
        )
        return res

    async def get_sub_deposit_records(
        self,
        subMemberId: str,
        coin: str = None,
        startTime: int = None,
        limit: int = 20,
    ):
        """
        :param subMemberId: str
        :param coin: str
        :param startTime: str
        :param limit: int
        """
        payload = {
            "subMemberId": subMemberId,
            "limit": limit,
        }
        if coin is not None:
            payload["coin"] = coin
        if startTime is not None:
            payload["startTime"] = startTime

        res = await self._request(
            method="GET",
            path=Asset.GET_SUB_ACCOUNT_DEPOSIT_RECORDS,
            query=payload,
        )
        return res

    async def get_internal_deposit_records(
        self,
        coin: str = None,
        startTime: int = None,
        limit: int = 20,
    ):
        """
        :param coin: str
        :param startTime: str
        :param limit: int
        """
        payload = {
            "limit": limit,
        }
        if coin is not None:
            payload["coin"] = coin
        if startTime is not None:
            payload["startTime"] = startTime

        res = await self._request(
            method="GET",
            path=Asset.GET_INTERNAL_DEPOSIT_RECORDS,
            query=payload,
        )
        return res

    async def get_master_deposit_address(
        self,
        coin: str,
    ):
        """
        :param coin: str
        """
        payload = {
            "coin": coin,
        }

        res = await self._request(
            method="GET",
            path=Asset.GET_MASTER_DEPOSIT_ADDRESS,
            query=payload,
        )
        return res

    async def get_withdrawal_records(
        self,
        coin: str = None,
        withdrawType: int = None,
        startTime: int = None,
        limit: int = 20,
    ):
        """
        :param coin: str
        :param withdrawType: int
        :param startTime: int
        :param limit: int
        """
        payload = {
            "limit": limit,
        }
        if coin is not None:
            payload["coin"] = coin
        if withdrawType is not None:
            payload["withdrawType"] = withdrawType
        if startTime is not None:
            payload["startTime"] = startTime

        res = await self._request(
            method="GET",
            path=Asset.GET_WITHDRAWAL_RECORDS,
            query=payload,
        )
        return res

    async def withdraw(
        self,
        coin: str,
        chain: str,
        address: str,
        amount: str,
        tag: str = None,
    ):
        """
        :param coin: str
        :param chain: str
        :param address: str
        :param amount: str
        :param tag: str
        """
        payload = {
            "coin": coin,
            "chain": chain,
            "address": address,
            "amount": amount,
            "timestamp": int(time.time() * 1000),
            "accountType": "FUND",
            "feeType": 1,
        }
        if chain is not None:
            payload["chain"] = chain
        if tag is not None:
            payload["tag"] = tag

        res = await self._request(
            method="POST",
            path=Asset.WITHDRAW,
            query=payload,
        )
        return res

    async def cancel_withdrawal(
        self,
        id: str,
    ):
        """
        :param id: str
        """
        payload = {
            "id": id,
        }

        res = await self._request(
            method="POST",
            path=Asset.CANCEL_WITHDRAWAL,
            query=payload,
        )
        return res



import httpx


class Slack:
    """Slack message sender that supports multiple bots and channels."""

    def __init__(self, bot_tokens: dict, channel_ids: dict):
        self.bot_tokens = bot_tokens  # e.g., {"GHOST": "xoxb-..."}
        self.channel_ids = channel_ids  # e.g., {"CRITICAL": "C1234", "ALL_KAIROS": "C5678"}

    async def send_message(self, bot_name: str, channel_name: str, text: str):
        bot_token = self.bot_tokens.get(bot_name)
        channel_id = self.channel_ids.get(channel_name)

        if not bot_token or not channel_id:
            raise ValueError(f"Invalid bot or channel name: {bot_name}, {channel_name}")

        url = "https://slack.com/api/chat.postMessage"
        headers = {"Authorization": f"Bearer {bot_token}"}
        data = {"channel": channel_id, "text": f"```{text}```"}

        async with httpx.AsyncClient() as client:
            response = await client.post(url, headers=headers, data=data)

        if response.status_code != 200 or not response.json().get("ok"):
            raise Exception(f"Error sending message: {response.text}")
        return response.json()

    async def send_file(self, bot_name: str, channel_name: str, file_path: str):
        bot_token = self.bot_tokens.get(bot_name)
        channel_id = self.channel_ids.get(channel_name)

        if not bot_token or not channel_id:
            raise ValueError(f"Invalid bot or channel name: {bot_name}, {channel_name}")

        url = "https://slack.com/api/files.upload"
        headers = {"Authorization": f"Bearer {bot_token}"}
        data = {"channels": channel_id}

        async with httpx.AsyncClient() as client:
            with open(file_path, "rb") as file:
                files = {"file": (file_path, file, "application/octet-stream")}
                response = await client.post(url, headers=headers, data=data, files=files)

        if response.status_code != 200 or not response.json().get("ok"):
            raise Exception(f"Error uploading file: {response.text}")
        return response.json()



from enum import Enum


class Market(str, Enum):
    GET_INSTRUMENTS_INFO = "/v5/market/instruments-info"
    GET_KLINE = "/v5/market/kline"
    GET_ORDERBOOK = "/v5/market/orderbook"
    GET_TICKERS = "/v5/market/tickers"
    GET_FUNDING_RATE_HISTORY = "/v5/market/funding/history"
    GET_PUBLIC_TRADE_HISTORY = "/v5/market/recent-trade"

    def __str__(self) -> str:
        return self.value



from enum import Enum


class SpotMarket(str, Enum):
    EXCHANGE_INFO = "/api/v3/exchangeInfo"

    def __str__(self) -> str:
        return self.value


class FuturesMarket(str, Enum):
    EXCHANGE_INFO = "/fapi/v1/exchangeInfo"
    BOOK_TICKER = "/fapi/v1/ticker/bookTicker"
    KLINE = "/fapi/v1/klines"
    PREMIUM_INDEX = "/fapi/v1/premiumIndex"
    FUNDING_RATE_HISTORY = "/fapi/v1/fundingRate"

    def __str__(self) -> str:
        return self.value



import hmac
import hashlib
import logging
import json
import requests
from dataclasses import dataclass, field
from ..product_table.manager import ProductTableManager
from ..utils.errors import FailedRequestError
from ..utils.helpers import generate_timestamp
from ..utils.common import Common

HTTP_URL = "https://{SUBDOMAIN}.{DOMAIN}.{TLD}"
SUBDOMAIN_MAIN = "api"
DOMAIN_MAINNET = "hyperliquid"
DOMAIN_TESTNET = "hyperliquid-testnet"
TLD_MAIN = "xyz"


def get_header_no_sign():
    return {"Content-Type": "application/json"}


@dataclass
class HTTPManager:
    testnet: bool = field(default=False)
    subdomain: str = field(default=SUBDOMAIN_MAIN)
    tld: str = field(default=TLD_MAIN)
    api_key: str = field(default=None)
    api_secret: str = field(default=None)
    timeout: int = field(default=10)
    recv_window: int = field(default=5000)
    max_retries: int = field(default=3)
    retry_delay: int = field(default=3)
    logger: logging.Logger = field(default=None)
    session: requests.Session = field(default_factory=requests.Session, init=False)
    ptm: ProductTableManager = field(init=False)
    preload_product_table: bool = field(default=True)

    def __post_init__(self):
        if self.logger is None:
            self._logger = logging.getLogger(__name__)
        else:
            self._logger = self.logger

        domain = DOMAIN_TESTNET if self.testnet else DOMAIN_MAINNET
        self.endpoint = HTTP_URL.format(SUBDOMAIN=self.subdomain, DOMAIN=domain, TLD=self.tld)

        if self.preload_product_table:
            self.ptm = ProductTableManager.get_instance(Common.HYPERLIQUID)

    def _auth(self, payload, timestamp):
        param_str = f"{timestamp}{self.api_key}{self.recv_window}{payload}"
        return hmac.new(self.api_secret.encode(), param_str.encode(), hashlib.sha256).hexdigest()

    def _request(
        self,
        method: str,
        path: str,
        query: dict = None,
        signed: bool = True,
    ):
        if query is None:
            query = {}

        timestamp = generate_timestamp()

        if method.upper() == "GET":
            if query:
                sorted_query = "&".join(f"{k}={v}" for k, v in sorted(query.items()) if v)
                path += "?" + sorted_query if sorted_query else ""
                payload = sorted_query
            else:
                payload = ""
        else:
            payload = json.dumps(query)

        # if signed:
        #     if not (self.api_key and self.api_secret):
        #         raise ValueError("Signed request requires API Key and Secret.")
        #     signature = self._auth(payload, timestamp)
        #     headers = get_header(self.api_key, signature, timestamp, self.recv_window)
        # else:
        headers = get_header_no_sign()

        url = self.endpoint + path

        try:
            if method.upper() == "GET":
                response = self.session.get(url, headers=headers, timeout=self.timeout)
            elif method.upper() == "POST":
                response = self.session.post(url, json=query if query else {}, headers=headers, timeout=self.timeout)
            else:
                raise ValueError(f"Unsupported HTTP method: {method}")

            try:
                data = response.json()
            except Exception:
                data = {}

            # If http status is not 2xx (like 403, 404)
            if not response.status_code // 100 == 2:
                raise FailedRequestError(
                    request=f"{method.upper()} {url} | Body: {query}",
                    message=f"HTTP Error {response.status_code}: {response.text}",
                    status_code=response.status_code,
                    time=timestamp,
                    resp_headers=response.headers,
                )

            return data

        except requests.exceptions.RequestException as e:
            raise FailedRequestError(
                request=f"{method.upper()} {url} | Body: {payload}",
                message=f"Request failed: {str(e)}",
                status_code=getattr(e.response, "status_code", "Unknown"),
                time=timestamp,
                resp_headers=getattr(e.response, "headers", None),
            )



from ._http_manager import HTTPManager
from .endpoints.position import Position
from ...utils.common import Common


class PositionHTTP(HTTPManager):
    async def get_positions(
        self,
        category: str = "linear",
        product_symbol: str = None,
        settleCoin: str = None,
        limit: int = 20,
    ):
        """
        :param category: str (linear, inverse, option)
        :param symbol: str
        :param limit: str
        """
        payload = {
            "category": category,
            "limit": limit,
        }
        if product_symbol is not None:
            payload["symbol"] = self.ptm.get_exchange_symbol(product_symbol, Common.BYBIT)
            payload["category"] = self.ptm.get_product_type(product_symbol, Common.BYBIT)
        if settleCoin is not None:
            payload["settleCoin"] = settleCoin

        res = await self._request(
            method="GET",
            path=Position.GET_POSITIONS,
            query=payload,
        )
        return res

    async def set_leverage(
        self,
        product_symbol: str,
        leverage: str,
    ):
        """
        :param category: str (linear, inverse)
        :param symbol: str
        :param buyLeverage: str
        :param sellLeverage: str
        """
        payload = {
            "category": self.ptm.get_product_type(product_symbol, Common.BYBIT),
            "symbol": self.ptm.get_exchange_symbol(product_symbol, Common.BYBIT),
            "buyLeverage": leverage,
            "sellLeverage": leverage,
        }

        res = await self._request(
            method="POST",
            path=Position.SET_LEVERAGE,
            query=payload,
        )
        return res

    async def switch_position_mode(
        self,
        mode: int,
        product_symbol: str = None,
        coin: str = None,
    ):
        """
        :param mode: int. 0: Merged Single. 3: Both Sides
        :param product_symbol: str
        :param coin: str
        """
        payload = {
            "category": "linear",
            "mode": mode,
        }
        if product_symbol is not None:
            payload["symbol"] = self.ptm.get_exchange_symbol(product_symbol, Common.BYBIT)
        if coin is not None:
            payload["coin"] = coin

        res = await self._request(
            method="POST",
            path=Position.SWITCH_POSITION_MODE,
            query=payload,
        )
        return res

    async def get_closed_pnl(
        self,
        category: str = "linear",
        product_symbol: str = None,
        startTime: int = None,
        limit: int = 20,
    ):
        """
        :param category: str
        :param symbol: str
        :param margin: str
        :param positionIdx: int
        """
        payload = {
            "category": category,
            "limit": limit,
        }
        if product_symbol is not None:
            payload["symbol"] = self.ptm.get_exchange_symbol(product_symbol, Common.BYBIT)
            payload["category"] = self.ptm.get_product_type(product_symbol, Common.BYBIT)
        if startTime is not None:
            payload["startTime"] = startTime

        res = await self._request(
            method="GET",
            path=Position.GET_CLOSED_PNL,
            query=payload,
        )
        return res



import asyncio
import websockets
import json
import logging
import traceback
from datetime import datetime
from .models.ticker import BookTicker

logger = logging.getLogger(__name__)
logging.basicConfig(
    level=logging.ERROR,
    format="%(asctime)s - %(levelname)s - %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


class WsClient:
    URI = ""
    SANDBOX_URI = ""

    def __init__(
        self,
        subscription: dict,
        is_sandbox: bool = False,
        orderbook_queue: asyncio.Queue = None,
        slack=None,
        slack_bot_name: str = None,
        slack_channel_name: str = None,
    ):
        self.uri = self.URI if not is_sandbox else self.SANDBOX_URI
        self.subscription = subscription
        self.orderbook_queue = orderbook_queue
        self.should_run = True
        self.websocket = None

        self.slack = slack
        self.slack_bot_name = slack_bot_name
        self.slack_channel_name = slack_channel_name

        self.message_check_task = None
        self.last_message_time = asyncio.get_event_loop().time()

    async def send_slack(self, msg: str):
        if self.slack and self.slack_bot_name and self.slack_channel_name:
            try:
                await self.slack.send_message(
                    self.slack_bot_name,
                    self.slack_channel_name,
                    f"\n{datetime.now().strftime('%Y-%m-%d %H:%M:%S')} - {msg}",
                )
            except Exception as e:
                logger.warning(f"Slack error: {e}")

    async def on_message(self, message: str):
        logger.info(f"Received message: {message}")

    async def update_orderbook_queue(self, exchange: str, symbol: str, ticker: BookTicker):
        if not self.orderbook_queue:
            return
        await self.orderbook_queue.put(
            (
                exchange,
                symbol,
                ticker,
            )
        )

    async def on_open(self):
        subscribe_message = json.dumps(self.subscription)
        await self.websocket.send(subscribe_message)
        await self.send_slack(f"[INFO] - {self.__class__.__name__} WebSocket connection opened")
        logger.info("WebSocket connection opened")
        logger.info(f"Sent subscription message: {subscribe_message}")

    async def on_close(self):
        logger.info("WebSocket connection closed")
        if self.message_check_task:
            self.message_check_task.cancel()
        await self.send_slack(f"[CRITICAL] - {self.__class__.__name__} WebSocket connection closed")

    async def on_error(self, error: Exception):
        await self.send_slack(f"[CRITICAL] - {self.__class__.__name__} Error occurred: {traceback.format_exc()}")
        await self.reconnect()

    async def message_check_loop(self):
        while self.should_run:
            now = asyncio.get_event_loop().time()
            if now - self.last_message_time > 30:
                await self.send_slack(
                    f"[CRITICAL] - {self.__class__.__name__} No message received for 30 seconds. Reconnecting..."
                )
                await self.reconnect()
                break
            await asyncio.sleep(1)

    async def receive_loop(self):
        first_message = True
        try:
            async for message in self.websocket:
                self.last_message_time = asyncio.get_event_loop().time()

                if first_message:
                    try:
                        parsed = json.loads(message)
                        print("📥 Received first raw message:")
                        print(json.dumps(parsed, indent=2, ensure_ascii=False))
                    except Exception:
                        print("📥 Received non-JSON first message:", message)
                    first_message = False

                await self.on_message(message)
        except Exception as e:
            await self.on_error(e)

    async def connect(self):
        async with websockets.connect(self.uri, ping_interval=10, ping_timeout=3) as ws:
            self.websocket = ws
            await self.on_open()

            try:
                self.message_check_task = asyncio.create_task(self.message_check_loop())
                await self.receive_loop()
            finally:
                await self.on_close()

    async def reconnect(self):
        if self.should_run:
            await self.connect()

    async def send_message(self, message: dict):
        if self.websocket:
            await self.websocket.send(json.dumps(message))
            logger.info(f"Sent message: {message}")
        else:
            logger.warning("WebSocket not connected")

    async def start(self):
        self.should_run = True
        while self.should_run:
            try:
                await self.connect()
            except Exception as e:
                logger.error(f"Connection error: {e}")
                await self.send_slack(
                    f"[CRITICAL] - {self.__class__.__name__}, Connection error: {traceback.format_exc()}"
                )
                await asyncio.sleep(0.1)

    async def stop(self):
        self.should_run = False
        if self.websocket:
            await self.websocket.close()



from ._http_manager import HTTPManager
from ._market_http import MarketHTTP
from .endpoint.path import Path
from .endpoint.trade import Trade
from ...utils.common import Common
import json


class TradeHTTP(HTTPManager):
    async def place_order(
        self,
        product_symbol: str,
        isBuy: bool,
        price: str,
        size: str,
        reduceOnly: bool,
        tif: str = None,
        isMarket: bool = None,
        triggerPx: str = None,
        tpsl: str = None,
        cloid: str = None,
        grouping: str = "na",
        builder_address: str = None,
        fee_ten_bp: int = None,
        vaultAddress: str = None,
        expireAfter: int = None,
    ):
        """
        :param product_symbol: str
        :param isBuy: bool
        :param price: str
        :param size: str
        :param reduceOnly: bool
        :param tif: str
        :param isMarket: bool
        :param triggerPx: str
        :param tpsl: str
        :param cloid: str
        :param grouping: str
        :param builder_address: str
        :param fee_ten_bp: int
        :param vaultAddress: str
        :param expireAfter: int
        """
        action = {
            "type": Trade.ORDER,
            "orders": [
                {
                    "a": json.loads(self.ptm.get_exchange_symbol(product_symbol, Common.HYPERLIQUID))[1],
                    "b": isBuy,
                    "p": price,
                    "s": size,
                    "r": reduceOnly,
                }
            ],
            "grouping": grouping,
        }

        if tif is not None or isMarket:
            t = {}
            if tif is not None:
                t["limit"] = {"tif": tif}
            else:
                t["trigger"] = {
                    "isMarket": isMarket,
                    "triggerPx": triggerPx,
                    "tpsl": tpsl,
                }
            action["orders"][0]["t"] = t

        if cloid is not None:
            action["orders"][0]["c"] = cloid
        if builder_address is not None:
            action["builder"]["b"] = builder_address
        if fee_ten_bp is not None:
            action["feeTenBp"] = fee_ten_bp

        payload = {
            "action": action,
            "nonce": "",
            "signature": "",
        }

        if vaultAddress is not None:
            payload["vaultAddress"] = vaultAddress
        if expireAfter is not None:
            payload["expireAfter"] = expireAfter

        res = await self._request(
            method="POST",
            path=Path.EXCHANGE,
            query=payload,
            signed=True,
        )
        return res

    async def place_future_market_order(
        self,
        product_symbol: str,
        isBuy: bool,
        size: str,
        triggerPx: str = None,
        tpsl: str = None,
    ):
        market_http = MarketHTTP()
        await market_http.async_init()
        result = await market_http.meta_and_asset_ctxs()
        exchange_symbol = int(self.ptm.get_exchange_symbol(product_symbol, Common.HYPERLIQUID))
        price = result[1][exchange_symbol]["midPx"]

        return await self.place_order(
            product_symbol=product_symbol,
            isBuy=isBuy,
            price=price.split(".", 1)[0],
            size=size,
            reduceOnly=False,
            isMarket=True,
            triggerPx=triggerPx,
            tpsl=tpsl,
        )

    async def place_future_market_buy_order(
        self,
        product_symbol: str,
        size: str,
        triggerPx: str = None,
        tpsl: str = None,
    ):
        return await self.place_future_market_order(
            product_symbol=product_symbol,
            isBuy=True,
            size=size,
            triggerPx=triggerPx,
            tpsl=tpsl,
        )

    async def place_future_market_sell_order(
        self,
        product_symbol: str,
        size: str,
        triggerPx: str = None,
        tpsl: str = None,
    ):
        return await self.place_future_market_order(
            product_symbol=product_symbol,
            isBuy=False,
            size=size,
            triggerPx=triggerPx,
            tpsl=tpsl,
        )

    async def place_future_limit_order(
        self,
        product_symbol: str,
        isBuy: bool,
        price,
        size: str,
        tif: str,
    ):
        return await self.place_order(
            product_symbol=product_symbol, isBuy=isBuy, price=price, size=size, reduceOnly=False, tif=tif
        )

    async def place_future_limit_buy_order(
        self,
        product_symbol: str,
        price,
        size: str,
        tif: str,
    ):
        return await self.place_order(
            product_symbol=product_symbol, isBuy=True, price=price, size=size, reduceOnly=False, tif=tif
        )

    async def place_future_limit_sell_order(
        self,
        product_symbol: str,
        price,
        size: str,
        tif: str,
    ):
        return await self.place_order(
            product_symbol=product_symbol, isBuy=False, price=price, size=size, reduceOnly=False, tif=tif
        )

    async def cancel_order(self, product_symbol: str, oid: int, vaultAddress: str = None, expireAfter: int = None):
        """
        :param product_symbol: str
        :param oid: str
        :param vaultAddress: str
        :param expireAfter: int
        """
        action = {
            "type": Trade.CANCEL,
            "cancels": [
                {
                    "a": json.loads(self.ptm.get_exchange_symbol(product_symbol, Common.HYPERLIQUID))[1],
                    "o": oid,
                }
            ],
        }

        payload = {
            "action": action,
            "nonce": "",
            "signature": "",
        }
        if vaultAddress is not None:
            payload["vaultAddress"] = vaultAddress
        if expireAfter is not None:
            payload["expireAfter"] = expireAfter

        res = await self._request(
            method="POST",
            path=Path.EXCHANGE,
            query=payload,
            signed=True,
        )
        return res

    async def cancel_order_by_cloid(
        self, product_symbol: str, cloid: str, vaultAddress: str = None, expireAfter: int = None
    ):
        """
        :param product_symbol: str
        :param cloid: str
        :param vaultAddress: str
        :param expireAfter: int
        """
        action = {
            "type": Trade.CANCELBYCLOID,
            "cancels": [
                {
                    "asset": json.loads(self.ptm.get_exchange_symbol(product_symbol, Common.HYPERLIQUID))[1],
                    "cloid": cloid,
                }
            ],
        }

        payload = {
            "action": action,
            "nonce": "",
            "signature": "",
        }
        if vaultAddress is not None:
            payload["vaultAddress"] = vaultAddress
        if expireAfter is not None:
            payload["expireAfter"] = expireAfter

        res = await self._request(
            method="POST",
            path=Path.EXCHANGE,
            query=payload,
            signed=True,
        )
        return res

    async def schedule_cancel(self, time: int = None, vaultAddress: str = None, expireAfter: int = None):
        """
        :param time: int
        :param vaultAddress: str
        :param expireAfter: int
        """
        action = {
            "type": Trade.SCHEDULECANCEL,
            "time": time,
        }

        payload = {
            "action": action,
            "nonce": "",
            "signature": "",
        }
        if vaultAddress is not None:
            payload["vaultAddress"] = vaultAddress
        if expireAfter is not None:
            payload["expireAfter"] = expireAfter

        res = await self._request(
            method="POST",
            path=Path.EXCHANGE,
            query=payload,
            signed=True,
        )
        return res

    async def modify_order(
        self,
        oid: int,
        product_symbol: str,
        isBuy: bool,
        price: str,
        size: str,
        reduceOnly: bool,
        tif: str = None,
        isMarket: bool = None,
        triggerPx: str = None,
        tpsl: str = None,
        cloid: str = None,
        vaultAddress: str = None,
        expireAfter: int = None,
    ):
        """
        :param oid: int
        :param product_symbol: str
        :param isBuy: bool
        :param price: str
        :param size: str
        :param reduceOnly: bool
        :param tif: str
        :param isMarket: bool
        :param triggerPx: str
        :param tpsl: str
        :param cloid: str
        :param vaultAddress: str
        :param expireAfter: int
        """
        action = {
            "type": Trade.MODIFY,
            "oid": oid,
            "order": {
                "a": json.loads(self.ptm.get_exchange_symbol(product_symbol, Common.HYPERLIQUID))[1],
                "b": isBuy,
                "p": price,
                "s": size,
                "r": reduceOnly,
            },
        }

        if tif is not None or isMarket:
            t = {}
            if tif is not None:
                t["limit"] = {"tif": tif}
            else:
                t["trigger"] = {
                    "isMarket": isMarket,
                    "triggerPx": triggerPx,
                    "tpsl": tpsl,
                }
            action["order"]["t"] = t
        if cloid is not None:
            action["order"]["c"] = cloid

        payload = {
            "action": action,
            "nonce": "",
            "signature": "",
        }

        if vaultAddress is not None:
            payload["vaultAddress"] = vaultAddress
        if expireAfter is not None:
            payload["expireAfter"] = expireAfter

        res = await self._request(
            method="POST",
            path=Path.EXCHANGE,
            query=payload,
            signed=True,
        )
        return res

    async def modify_batch_orders(self, modifies: list, vaultAddress: str = None, expireAfter: int = None):
        """
        :param modifies: list
        :param vaultAddress: str
        :param expireAfter: int
        """
        action = {"type": Trade.BATCHMODIFY, "modifies": modifies}

        payload = {
            "action": action,
            "nonce": "",
            "signature": "",
        }

        if vaultAddress is not None:
            payload["vaultAddress"] = vaultAddress
        if expireAfter is not None:
            payload["expireAfter"] = expireAfter

        res = await self._request(
            method="POST",
            path=Path.EXCHANGE,
            query=payload,
            signed=True,
        )
        return res

    async def update_leverage(
        self, product_symbol: str, isCross: bool, leverage: int, vaultAddress: str = None, expireAfter: int = None
    ):
        """
        :param product_symbol: str
        :param isCross: bool
        :param leverage: int
        :param vaultAddress: str
        :param expireAfter: int
        """
        action = {
            "type": Trade.UPDATELEVERAGE,
            "asset": json.loads(self.ptm.get_exchange_symbol(product_symbol, Common.HYPERLIQUID))[1],
            "isCross": isCross,
            "leverage": leverage,
        }

        payload = {
            "action": action,
            "nonce": "",
            "signature": "",
        }
        if vaultAddress is not None:
            payload["vaultAddress"] = vaultAddress
        if expireAfter is not None:
            payload["expireAfter"] = expireAfter

        res = await self._request(
            method="POST",
            path=Path.EXCHANGE,
            query=payload,
            signed=True,
        )
        return res

    async def update_isolate_margin(
        self, product_symbol: str, isBuy: bool, ntli: int, vaultAddress: str = None, expireAfter: int = None
    ):
        """
        :param product_symbol: str
        :param isBuy: bool
        :param ntli: int
        :param vaultAddress: str
        :param expireAfter: int
        """
        action = {
            "type": Trade.UPDATEISOLATEMARGIN,
            "asset": json.loads(self.ptm.get_exchange_symbol(product_symbol, Common.HYPERLIQUID))[1],
            "isBuy": isBuy,
            "ntli": ntli,
        }

        payload = {
            "action": action,
            "nonce": "",
            "signature": "",
        }
        if vaultAddress is not None:
            payload["vaultAddress"] = vaultAddress
        if expireAfter is not None:
            payload["expireAfter"] = expireAfter

        res = await self._request(
            method="POST",
            path=Path.EXCHANGE,
            query=payload,
            signed=True,
        )
        return res

    async def place_twap_order(
        self,
        product_symbol: str,
        isBuy: bool,
        size: str,
        reduceOnly: bool,
        minutes: int,
        randomize: bool,
        vaultAddress: str = None,
        expireAfter: int = None,
    ):
        """
        :param product_symbol: str
        :param isBuy: bool
        :param size: str
        :param reduceOnly: bool
        :param minutes: int
        :param randomize: bool
        :param vaultAddress: str
        :param expireAfter: int
        """
        action = {
            "type": Trade.TWAPORDER,
            "twap": {
                "a": json.loads(self.ptm.get_exchange_symbol(product_symbol, Common.HYPERLIQUID))[1],
                "b": isBuy,
                "s": size,
                "r": reduceOnly,
                "m": minutes,
                "t": randomize,
            },
        }

        payload = {
            "action": action,
            "nonce": "",
            "signature": "",
        }

        if vaultAddress is not None:
            payload["vaultAddress"] = vaultAddress
        if expireAfter is not None:
            payload["expireAfter"] = expireAfter

        res = await self._request(
            method="POST",
            path=Path.EXCHANGE,
            query=payload,
            signed=True,
        )
        return res

    async def cancel_twap_order(
        self, product_symbol: str, twap_id: int, vaultAddress: str = None, expireAfter: int = None
    ):
        """
        :param product_symbol: str
        :param twap_id: int
        :param vaultAddress: str
        :param expireAfter: int
        """
        action = {
            "type": Trade.TWAPCANCEL,
            "a": json.loads(self.ptm.get_exchange_symbol(product_symbol, Common.HYPERLIQUID))[1],
            "t": twap_id,
        }

        payload = {
            "action": action,
            "nonce": "",
            "signature": "",
        }

        if vaultAddress is not None:
            payload["vaultAddress"] = vaultAddress
        if expireAfter is not None:
            payload["expireAfter"] = expireAfter

        res = await self._request(
            method="POST",
            path=Path.EXCHANGE,
            query=payload,
            signed=True,
        )
        return res



def bybit_convert_timeframe(timeframe: str) -> int:
    if timeframe == "1m":
        return "1"
    elif timeframe == "3m":
        return "3"
    elif timeframe == "5m":
        return "5"
    elif timeframe == "15m":
        return "15"
    elif timeframe == "30m":
        return "30"
    elif timeframe == "1h":
        return "60"
    elif timeframe == "2h":
        return "120"
    elif timeframe == "4h":
        return "240"
    elif timeframe == "6h":
        return "360"
    elif timeframe == "12h":
        return "720"
    elif timeframe == "1d":
        return "D"
    elif timeframe == "1w":
        return "W"
    elif timeframe == "1M":
        return "M"


def bitmart_convert_timeframe(timeframe: str) -> str:
    if timeframe == "1m":
        return 1
    elif timeframe == "5m":
        return 5
    elif timeframe == "15m":
        return 15
    elif timeframe == "30m":
        return 30
    elif timeframe == "1h":
        return 60
    elif timeframe == "2h":
        return 120
    elif timeframe == "4h":
        return 240
    elif timeframe == "1d":
        return 1440
    elif timeframe == "1w":
        return 10080
    elif timeframe == "1mo":
        return 43200
    else:
        raise ValueError("timeframe not supported")



from dataclasses import dataclass
from enum import Enum


class OrderStatus(str, Enum):
    OPEN = "open"
    FILLED = "filled"
    CANCELLED = "cancelled"
    PARTIALLY_FILLED = "partially_filled"
    # REJECTED = "rejected"


class OrderType(str, Enum):
    LIMIT = "limit"
    MARKET = "market"


class OrderMode(str, Enum):
    POST_ONLY = "post_only"
    IOC = "ioc"
    FOK = "fok"
    GTC = "gtc"


class OrderSide(str, Enum):
    BUY = "buy"
    SELL = "sell"


class Exchange(str, Enum):
    OKX = "okx"
    BITMART = "bitmart"
    BYBIT = "bybit"


@dataclass
class TradeStatus:
    order_id: str
    client_oid: str
    order_type: str
    status: OrderStatus
    price: float
    quantity: float
    filled_quantity: float
    deal_avg_price: float


@dataclass
class TradeRequest:
    exchange: Exchange
    symbol: str
    side: OrderSide
    price: float
    quantity: float
    order_type: OrderType
    mode: OrderMode
    client_oid: str
    is_open_new_position: bool = False
    create_time: int = None



from ._http_manager import HTTPManager
from .endpoints.market import Market
from ..utils.common import Common
from ..utils.timeframe_utils import bybit_convert_timeframe


class MarketHTTP(HTTPManager):
    def get_instruments_info(
        self,
        category: str = "linear",
        product_symbol: str = None,
        status: str = None,
        baseCoin: str = None,
        limit: int = None,
    ):
        """
        :param category: str (spot, linear, inverse, option)
        :param product_symbol: str
        :param status: str
        :param baseCoin: str
        :param limit: int
        """
        payload = {
            "category": category,
        }
        if product_symbol is not None:
            payload["symbol"] = self.ptm.get_exchange_symbol(product_symbol, Common.BYBIT)
            payload["category"] = self.ptm.get_product_type(product_symbol, Common.BYBIT)
        if status is not None:
            payload["status"] = status
        if baseCoin is not None:
            payload["baseCoin"] = baseCoin
        if limit is not None:
            payload["limit"] = limit

        res = self._request(
            method="GET",
            path=Market.GET_INSTRUMENTS_INFO,
            query=payload,
            signed=False,
        )
        return res

    def get_kline(
        self,
        product_symbol: str,
        interval: str,
        startTime: int = None,
        limit: int = None,
    ):
        """
        :param symbol: str
        :param interval: str
        :param category: str (spot, linear, inverse) default is linear
        :param startTime: int
        :param limit: int
        """
        payload = {
            "symbol": self.ptm.get_exchange_symbol(product_symbol, Common.BYBIT),
            "category": self.ptm.get_product_type(product_symbol, Common.BYBIT),
            "interval": bybit_convert_timeframe(interval),
        }
        if startTime is not None:
            payload["start"] = startTime
        if limit is not None:
            payload["limit"] = limit

        res = self._request(
            method="GET",
            path=Market.GET_KLINE,
            query=payload,
            signed=False,
        )
        return res

    def get_orderbook(
        self,
        product_symbol: str,
        limit: int = None,
    ):
        """
        :param category: str (linear, inverse)
        :param symbol: str
        :param limit: int
        """
        payload = {
            "category": self.ptm.get_product_type(product_symbol, Common.BYBIT),
            "symbol": self.ptm.get_exchange_symbol(product_symbol, Common.BYBIT),
        }
        if limit is not None:
            payload["limit"] = limit

        res = self._request(
            method="GET",
            path=Market.GET_ORDERBOOK,
            query=payload,
            signed=False,
        )
        return res

    def get_tickers(
        self,
        category: str = "linear",
        product_symbol: str = None,
        baseCoin: str = None,
    ):
        """
        :param category: str (spot, linear, inverse, option)
        :param symbol: str
        :param baseCoin: str
        """
        payload = {
            "category": category,
        }
        if product_symbol is not None:
            payload["symbol"] = self.ptm.get_exchange_symbol(product_symbol, Common.BYBIT)
            payload["category"] = self.ptm.get_product_type(product_symbol, Common.BYBIT)
        if baseCoin is not None:
            payload["baseCoin"] = baseCoin

        res = self._request(
            method="GET",
            path=Market.GET_TICKERS,
            query=payload,
            signed=False,
        )
        return res

    def get_funding_rate_history(
        self,
        product_symbol: str,
        startTime: int = None,
        limit: int = None,
    ):
        """
        :param category: str (linear, inverse)
        :param symbol: str
        :param startTime: int
        :param limit: int
        """
        payload = {
            "category": self.ptm.get_product_type(product_symbol, Common.BYBIT),
            "symbol": self.ptm.get_exchange_symbol(product_symbol, Common.BYBIT),
        }
        if startTime is not None:
            payload["startTime"] = startTime
        if limit is not None:
            payload["limit"] = limit

        res = self._request(
            method="GET",
            path=Market.GET_FUNDING_RATE_HISTORY,
            query=payload,
            signed=False,
        )
        return res

    def get_public_trade_history(
        self,
        product_symbol: str,
        limit: int = None,
    ):
        """
        :param category: str (linear, spot)
        :param symbol: str
        :param limit: int
        """
        payload = {
            "category": self.ptm.get_product_type(product_symbol, Common.BYBIT),
            "symbol": self.ptm.get_exchange_symbol(product_symbol, Common.BYBIT),
        }
        if limit is not None:
            payload["limit"] = limit

        res = self._request(
            method="GET",
            path=Market.GET_PUBLIC_TRADE_HISTORY,
            query=payload,
            signed=False,
        )
        return res



import asyncio
import polars as pl
from typing import Dict, Any

from ..models.order import TradeStatus
from ..models.position import PositionInfo
from ..models.ticker import BookTicker
from ..models.balance import BalanceInfo


class MarketData:
    _instance = None
    _lock = asyncio.Lock()

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(MarketData, cls).__new__(cls)
            cls._instance._initialized = False
        return cls._instance

    def __init__(self):
        self._initialized = False
        self._data_lock = asyncio.Lock()
        self._depth_data = {}
        self._kline_data = {}
        self._position_data = {}
        self._trade_data = {}
        self._balance_data = {}

    async def update_depth_data(self, exchange: str, symbol: str, data: BookTicker) -> None:
        async with self._data_lock:
            self._depth_data.setdefault(exchange, {})[symbol] = data

    async def get_depth_data(self, exchange: str, symbol: str) -> Any:
        async with self._data_lock:
            return self._depth_data.get(exchange, {}).get(symbol)

    async def get_all_depth_data(self) -> Dict[str, Any]:
        async with self._data_lock:
            return self._depth_data.copy()

    async def update_kline_data(self, exchange: str, symbol: str, data: pl.DataFrame) -> None:
        async with self._data_lock:
            self._kline_data.setdefault(exchange, {})
            if symbol in self._kline_data[exchange]:
                existing_df = self._kline_data[exchange][symbol]
                combined_df = pl.concat([existing_df, data])
            else:
                combined_df = data

            combined_df = combined_df.sort("datetime")
            combined_df = combined_df.unique(subset=["datetime"], keep="last")
            if combined_df.height > 1000:
                combined_df = combined_df.slice(-1000, 1000)

            self._kline_data[exchange][symbol] = combined_df

    async def get_kline_data(self, exchange: str, symbol: str) -> Any:
        async with self._data_lock:
            return self._kline_data.get(exchange, {}).get(symbol)

    async def get_all_kline_data(self) -> Dict[str, Any]:
        async with self._data_lock:
            return self._kline_data.copy()

    async def update_position(self, exchange: str, symbol: str, data: PositionInfo) -> None:
        async with self._data_lock:
            self._position_data.setdefault(exchange, {})[symbol] = data

    async def get_position(self, symbol: str) -> Any:
        async with self._data_lock:
            return self._position_data.get(symbol)

    async def get_all_positions(self) -> Dict[str, Any]:
        async with self._data_lock:
            return self._position_data.copy()

    async def update_trade_status(self, exchange: str, symbol: str, data: TradeStatus) -> None:
        async with self._data_lock:
            self._trade_data.setdefault(exchange, {})[symbol] = data

    async def get_trade_status(self, order_id: str) -> Any:
        async with self._data_lock:
            return self._trade_data.get(order_id)

    async def get_all_trade_statuses(self) -> Dict[str, Any]:
        async with self._data_lock:
            return self._trade_data.copy()

    async def update_balance(self, exchange: str, symbol: str, data: BalanceInfo) -> None:
        async with self._data_lock:
            self._balance_data.setdefault(exchange, {})[symbol] = data

    async def get_balance(self, symbol: str) -> Any:
        async with self._data_lock:
            return self._balance_data.get(symbol)

    async def get_all_balances(self) -> Dict[str, Any]:
        async with self._data_lock:
            return self._balance_data.copy()



from enum import Enum


class SpotMarket(str, Enum):
    # https://api-cloud.bitmart.com
    GET_SPOT_CURRENCIES = "/spot/v1/currencies"
    GET_TRADING_PAIRS = "/spot/v1/symbols"
    GET_TRADING_PAIRS_DETAILS = "/spot/v1/symbols/details"
    GET_TICKER_OF_ALL_PAIRS = "/spot/quotation/v3/tickers"
    GET_TICKER_OF_A_PAIR = "/spot/quotation/v3/ticker"
    GET_SPOT_KLINE = "/spot/quotation/v3/lite-klines"

    def __str__(self) -> str:
        return self.value


class FuturesMarket(str, Enum):
    # https://api-cloud-v2.bitmart.com
    GET_CONTRACTS_DETAILS = "/contract/public/details"
    GET_DEPTH = "/contract/public/depth"
    GET_CONTRACTS_KLINE = "/contract/public/kline"
    GET_CURRENT_FUNDING_RATE = "/contract/public/funding-rate"
    GET_FUNDING_RATE_HISTORY = "/contract/public/funding-rate-history"

    def __str__(self) -> str:
        return self.value



import json
import time
import hmac
import logging
import asyncio
import polars as pl
from .base import WsClient
from .data.market_data import MarketData
from .models.balance import BalanceInfo
from .models.order import Exchange, OrderStatus, TradeStatus
from .models.ticker import BookTicker
from .models.position import PositionInfo, PositionSide

logger = logging.getLogger(__name__)


def safe_float(value, default=0.0):
    if value is None or value == "":
        return default
    try:
        return float(value)
    except (ValueError, TypeError):
        return default


class BybitPublicLinearWsClient(WsClient):
    URI = "wss://stream.bybit.com/v5/public/linear"
    SANDBOX_URI = "wss://stream-testnet.bybit.com/v5/public/linear"

    def __init__(
        self,
        subscription: dict,
        is_sandbox: bool = False,
        orderbook_queue=None,
        slack=None,
        slack_bot_name: str = None,
        slack_channel_name: str = None,
    ):
        super().__init__(subscription, is_sandbox, orderbook_queue, slack, slack_bot_name, slack_channel_name)
        self.market_data = MarketData()

    @classmethod
    async def create(cls, **kwargs):
        self = cls(**kwargs)
        return self

    async def on_open(self):
        linear_args_plan = [self.subscription["args"][i : i + 10] for i in range(0, len(self.subscription["args"]), 10)]
        for linear_args in linear_args_plan:
            subscribe_message = json.dumps(
                {
                    "op": "subscribe",
                    "args": linear_args,
                }
            )
            await self.websocket.send(subscribe_message)
            logger.info(f"Sent subscription message: {subscribe_message}")
            await asyncio.sleep(0.3)
        await self.send_slack(f"[INFO] - {self.__class__.__name__} WebSocket connection opened")

    async def on_message(self, message: str):
        await super().on_message(message)
        payload = json.loads(message)

        topic = payload.get("topic", "")
        data = payload.get("data", [])

        if topic.startswith("orderbook.1.") and data:
            symbol = topic.split(".")[-1]
            book_ticker = BookTicker(
                symbol=symbol,
                product_type="linear",
                best_bid_price=float(data["b"][0][0]),
                best_bid_quantity=float(data["b"][0][1]),
                best_ask_price=float(data["a"][0][0]),
                best_ask_quantity=float(data["a"][0][1]),
                timestamp=int(payload.get("ts")),
            )
            await self.market_data.update_depth_data("bybit", symbol, book_ticker)
            await self.update_orderbook_queue("bybit", symbol, book_ticker)

        elif topic.startswith("tickers.") and data:
            symbol = topic.split(".")[-1]
            ticker_data = data[0]
            book_ticker = BookTicker(
                symbol=symbol,
                product_type="linear",
                best_bid_price=float(ticker_data["bid1Price"]),
                best_bid_quantity=float(ticker_data["bid1Size"]),
                best_ask_price=float(ticker_data["ask1Price"]),
                best_ask_quantity=float(ticker_data["ask1Size"]),
                timestamp=int(payload.get("ts")),
            )

        elif topic.startswith("kline.") and data:
            kline = data[0]
            symbol = topic.split(".")[-1]

            df = (
                pl.DataFrame(
                    {
                        "open": [float(kline["open"])],
                        "high": [float(kline["high"])],
                        "low": [float(kline["low"])],
                        "close": [float(kline["close"])],
                        "volume": [float(kline["volume"])],
                        "datetime": [int(kline["start"])],
                    }
                )
                .with_columns(pl.col("datetime").cast(pl.Int64).cast(pl.Datetime("ms")))
                .set_sorted("datetime")
            )
            await self.market_data.update_kline_data("bybit", symbol, df)


class BybitPublicSpotWsClient(WsClient):
    URI = "wss://stream.bybit.com/v5/public/spot"
    SANDBOX_URI = "wss://stream-testnet.bybit.com/v5/public/spot"

    def __init__(
        self,
        subscription: dict,
        is_sandbox: bool = False,
        orderbook_queue=None,
        slack=None,
        slack_bot_name: str = None,
        slack_channel_name: str = None,
    ):
        super().__init__(subscription, is_sandbox, orderbook_queue, slack, slack_bot_name, slack_channel_name)
        self.market_data = MarketData()

    @classmethod
    async def create(cls, **kwargs):
        self = cls(**kwargs)
        return self

    async def on_open(self):
        spot_args_plan = [self.subscription["args"][i : i + 10] for i in range(0, len(self.subscription["args"]), 10)]
        for spot_args in spot_args_plan:
            subscribe_message = json.dumps(
                {
                    "op": "subscribe",
                    "args": spot_args,
                }
            )
            await self.websocket.send(subscribe_message)
            logger.info(f"Sent subscription message: {subscribe_message}")
            await asyncio.sleep(0.3)
        await self.send_slack(f"[INFO] - {self.__class__.__name__} WebSocket connection opened")

    async def on_message(self, message: str):
        await super().on_message(message)
        payload = json.loads(message)

        topic = payload.get("topic", "")
        data = payload.get("data", [])

        if topic.startswith("orderbook.1.") and data:
            symbol = topic.split(".")[-1]
            book_ticker = BookTicker(
                symbol=symbol,
                product_type="spot",
                best_bid_price=float(data["b"][0][0]),
                best_bid_quantity=float(data["b"][0][1]),
                best_ask_price=float(data["a"][0][0]),
                best_ask_quantity=float(data["a"][0][1]),
                timestamp=int(payload.get("ts")),
            )
            await self.market_data.update_depth_data("bybit", symbol, book_ticker)
            await self.update_orderbook_queue("bybit", symbol, book_ticker)

        elif topic.startswith("tickers.") and data:
            symbol = topic.split(".")[-1]
            ticker_data = data[0]
            book_ticker = BookTicker(
                symbol=symbol,
                product_type="spot",
                best_bid_price=float(ticker_data["bid1Price"]),
                best_bid_quantity=float(ticker_data["bid1Size"]),
                best_ask_price=float(ticker_data["ask1Price"]),
                best_ask_quantity=float(ticker_data["ask1Size"]),
                timestamp=int(payload.get("ts")),
            )
            await self.market_data.update_depth_data("bybit", symbol, book_ticker)

        elif topic.startswith("kline.") and data:
            kline = data[0]
            symbol = topic.split(".")[-1]

            df = (
                pl.DataFrame(
                    {
                        "open": [float(kline["open"])],
                        "high": [float(kline["high"])],
                        "low": [float(kline["low"])],
                        "close": [float(kline["close"])],
                        "volume": [float(kline["volume"])],
                        "datetime": [int(kline["start"])],
                    }
                )
                .with_columns(pl.col("datetime").cast(pl.Int64).cast(pl.Datetime("ms")))
                .set_sorted("datetime")
            )
            await self.market_data.update_kline_data("bybit", symbol, df)


class BybitPrivateWsClient(WsClient):
    URI = "wss://stream.bybit.com/v5/private"
    SANDBOX_URI = "wss://stream-testnet.bybit.com/v5/private"

    def __init__(
        self,
        subscription: dict,
        api_key: str,
        api_secret: str,
        is_sandbox: bool = False,
        slack=None,
        slack_bot_name: str = None,
        slack_channel_name: str = None,
    ):
        super().__init__(subscription, is_sandbox, slack, slack_bot_name, slack_channel_name)
        self.api_key = api_key
        self.api_secret = api_secret
        self.market_data = MarketData()

    @classmethod
    async def create(cls, **kwargs):
        self = cls(**kwargs)
        return self

    def sign(self):
        expires = str(int((time.time() + 60) * 1000))
        message = "GET/realtime" + expires
        signature = hmac.new(self.api_secret.encode("utf-8"), message.encode("utf-8"), digestmod="sha256").hexdigest()
        return {
            "op": "auth",
            "args": [self.api_key, int(expires), signature],
        }

    async def on_open(self):
        await super().on_open()

        sign_message = json.dumps(self.sign())
        await self.websocket.send(sign_message)

        while True:
            message = await self.websocket.recv()
            payload = json.loads(message)
            if payload.get("op") == "auth" and payload.get("success"):
                break
            logger.debug(f"Waiting for auth confirmation: {payload}")

        subscribe_message = json.dumps(self.subscription)
        await self.websocket.send(subscribe_message)
        logger.info(f"Sent subscription message: {subscribe_message}")

    async def on_message(self, message: str):
        await super().on_message(message)
        payload = json.loads(message)

        topic = payload.get("topic", "")
        data = payload.get("data", [])

        if topic == "order" and data:
            for order in data:
                try:
                    trade_status = TradeStatus(
                        order_id=order["orderId"],
                        client_oid=order.get("orderLinkId", ""),
                        order_type=order.get("orderType", ""),
                        status=self.map_order_status(order.get("orderStatus", "")),
                        price=safe_float(order.get("price")),
                        quantity=safe_float(order.get("qty")),
                        filled_quantity=safe_float(order.get("cumExecQty")),
                        deal_avg_price=safe_float(order.get("avgPrice")),
                    )

                    await self.market_data.update_trade_status(Exchange.BYBIT, order["symbol"], trade_status)
                    logger.info(f"Updated TradeStatus: {trade_status}")
                except Exception as e:
                    logger.error(f"Error processing order data: {e}")

        elif topic == "position" and data:
            for position in data:
                try:
                    side = PositionSide.LONG if position.get("side") == "Buy" else PositionSide.SHORT
                    pos_info = PositionInfo(
                        symbol=position["symbol"],
                        side=side,
                        quantity=safe_float(position.get("size")),
                        open_avg_price=safe_float(position.get("entryPrice")),
                    )
                    await self.market_data.update_position(Exchange.BYBIT, position["symbol"], pos_info)
                    logger.info(f"Updated PositionInfo: {pos_info}")
                except Exception as e:
                    logger.error(f"Error processing position data: {e}")

        elif topic == "wallet" and data:
            try:
                wallet = data[0]
                for coin in wallet.get("coin", []):
                    balance_info = BalanceInfo(
                        symbol=coin["coin"],
                        available_balance=safe_float(coin.get("walletBalance")),
                        position_deposit=0.0,
                        frozen_balance=0.0,
                    )
                    await self.market_data.update_balance(Exchange.BYBIT, coin["coin"], balance_info)
                    logger.info(f"Updated BalanceInfo: {balance_info}")
            except Exception as e:
                logger.error(f"Error processing wallet data: {e}")

    def map_order_status(self, status: str):
        mapping = {
            "Created": OrderStatus.OPEN,
            "New": OrderStatus.OPEN,
            "Filled": OrderStatus.FILLED,
            "Cancelled": OrderStatus.CANCELLED,
            "Rejected": OrderStatus.CANCELLED,
            "PartiallyFilled": OrderStatus.PARTIALLY_FILLED,
        }
        return mapping.get(status, OrderStatus.OPEN)



from enum import Enum


class Trade(str, Enum):
    PLACE_ORDER = "/v5/order/create"
    AMEND_ORDER = "/v5/order/amend"
    CANCEL_ORDER = "/v5/order/cancel"
    GET_OPEN_ORDERS = "/v5/order/realtime"
    CANCEL_BATCH_ORDERS = "/v5/order/cancel-batch"
    CANCEL_ALL_ORDERS = "/v5/order/cancel-all"
    GET_ORDER_HISTORY = "/v5/order/history"
    GET_EXECUTION_LIST = "/v5/execution/list"
    BATCH_PLACE_ORDER = "/v5/order/create-batch"
    BATCH_AMEND_ORDER = "/v5/order/amend-batch"
    GET_BORROW_QUOTA = "/v5/order/spot-borrow-check"

    def __str__(self) -> str:
        return self.value


class SpotMarginTrade(str, Enum):
    VIP_MARGIN_DATA = "/v5/spot-margin-trade/data"
    GET_COLLATERAL = "/v5/spot-margin-trade/collateral"
    HISTORICAL_INTEREST = "/v5/spot-margin-trade/interest-rate-history"
    TOGGLE_MARGIN_TRADE = "/v5/spot-margin-trade/switch-mode"
    SET_LEVERAGE = "/v5/spot-margin-trade/set-leverage"
    STATUS_AND_LEVERAGE = "/v5/spot-margin-trade/state"

    def __str__(self) -> str:
        return self.value



"""Helper functions for Jupyter Notebook compatibility."""

import asyncio
import sys
from typing import Optional


def is_jupyter_environment() -> bool:
    """
    檢測是否在 Jupyter 環境中運行
    
    Returns:
        bool: True if running in Jupyter, False otherwise
    """
    try:
        # 檢查是否有 IPython 核心
        from IPython import get_ipython
        ipython = get_ipython()
        
        if ipython is None:
            return False
            
        # 檢查是否在 notebook 或 lab 環境中
        if hasattr(ipython, 'kernel'):
            return True
            
        # 檢查 IPython 的類名
        class_name = ipython.__class__.__name__
        if 'ZMQ' in class_name or 'Terminal' in class_name:
            return True
            
        return False
    except ImportError:
        return False


def has_running_event_loop() -> bool:
    """
    檢查是否已經有運行中的事件循環
    
    Returns:
        bool: True if there's a running event loop, False otherwise
    """
    try:
        loop = asyncio.get_running_loop()
        return loop is not None
    except RuntimeError:
        return False


def ensure_nest_asyncio() -> Optional[str]:
    """
    自動偵測 Jupyter 環境並應用 nest_asyncio
    
    Returns:
        Optional[str]: 返回狀態訊息，如果不需要應用則返回 None
    """
    # 如果不在 Jupyter 環境中，不需要 nest_asyncio
    if not is_jupyter_environment():
        return None
    
    # 如果沒有運行中的事件循環，也不需要 nest_asyncio
    if not has_running_event_loop():
        return None
    
    try:
        import nest_asyncio
        
        # 檢查是否已經應用過
        if hasattr(asyncio, '_nest_patched'):
            return "nest_asyncio already applied"
        
        # 應用 nest_asyncio
        nest_asyncio.apply()
        
        return "nest_asyncio applied successfully for Jupyter environment"
        
    except ImportError:
        warning_msg = (
            "Warning: Running in Jupyter environment with active event loop, "
            "but nest_asyncio is not installed. "
            "Please install it with: pip install nest_asyncio"
        )
        print(warning_msg, file=sys.stderr)
        return warning_msg


def auto_apply_nest_asyncio(verbose: bool = False) -> None:
    """
    自動應用 nest_asyncio（如果需要的話）
    
    Args:
        verbose (bool): 是否顯示詳細訊息
    """
    result = ensure_nest_asyncio()
    
    if verbose and result:
        print(f"[krex] {result}")


# 在模組導入時自動檢查並應用
auto_apply_nest_asyncio(verbose=False) 


from .endpoints.trade import FuturesTrade
from ._http_manager import HTTPManager
from ..utils.common import Common


class TradeHTTP(HTTPManager):
    def set_leverage(
        self,
        product_symbol: str,
        leverage: int,
    ):
        """
        :param product_symbol: str
        :param leverage: int
        """
        payload = {
            "symbol": self.ptm.get_exchange_symbol(product_symbol, Common.BINANCE),
            "leverage": leverage,
        }

        res = self._request(
            method="POST",
            path=FuturesTrade.SET_LEVERAGE,
            query=payload,
        )
        return res

    def place_future_order(
        self,
        product_symbol: str,
        side: str,
        type_: str,
        quantity: str = None,
        price: str = None,
        timeInForce: str = None,
        positionSide: str = None,
        reduceOnly: str = None,
        stopPrice: str = None,
        closePosition: str = None,
        activationPrice: str = None,
        callbackRate: str = None,
        workingType: str = None,
        priceProtect: str = None,
        newClientOrderId: str = None,
        newOrderRespType: str = None,
        priceMatch: str = None,
        selfTradePreventionMode: str = None,
        goodTillDate: int = None,
    ):
        """
        :param product_symbol: str
        :param side: str
        :param type_: str
        :param quantity: str
        :param price: str
        :param timeInForce: str
        :param positionSide: str
        :param reduceOnly: str
        :param stopPrice: str
        :param closePosition: str
        :param activationPrice: str
        :param callbackRate: str
        :param workingType: str
        :param priceProtect: str
        :param newClientOrderId: str
        :param newOrderRespType: str
        :param priceMatch: str
        :param selfTradePreventionMode: str
        :param goodTillDate: int
        """
        payload = {
            "symbol": self.ptm.get_exchange_symbol(product_symbol, Common.BINANCE),
            "side": side,
            "type": type_,
        }

        if quantity is not None:
            payload["quantity"] = quantity
        if price is not None:
            payload["price"] = price
        if timeInForce is not None:
            payload["timeInForce"] = timeInForce
        if positionSide is not None:
            payload["positionSide"] = positionSide
        if reduceOnly is not None:
            payload["reduceOnly"] = reduceOnly
        if stopPrice is not None:
            payload["stopPrice"] = stopPrice
        if closePosition is not None:
            payload["closePosition"] = closePosition
        if activationPrice is not None:
            payload["activationPrice"] = activationPrice
        if callbackRate is not None:
            payload["callbackRate"] = callbackRate
        if workingType is not None:
            payload["workingType"] = workingType
        if priceProtect is not None:
            payload["priceProtect"] = priceProtect
        if newClientOrderId is not None:
            payload["newClientOrderId"] = newClientOrderId
        if newOrderRespType is not None:
            payload["newOrderRespType"] = newOrderRespType
        if priceMatch is not None:
            payload["priceMatch"] = priceMatch
        if selfTradePreventionMode is not None:
            payload["selfTradePreventionMode"] = selfTradePreventionMode
        if goodTillDate is not None:
            payload["goodTillDate"] = goodTillDate

        res = self._request(
            method="POST",
            path=FuturesTrade.PLACE_CANCEL_QUERY_ORDER,
            query=payload,
        )
        return res

    def place_future_market_order(
        self,
        product_symbol: str,
        side: str,
        quantity: str,
        positionSide: str = None,
        reduceOnly: str = None,
    ):
        return self.place_future_order(
            product_symbol=product_symbol,
            side=side,
            type_="MARKET",
            quantity=quantity,
            positionSide=positionSide,
            reduceOnly=reduceOnly,
        )

    def place_future_market_buy_order(
        self,
        product_symbol: str,
        quantity: str,
        positionSide: str = None,
        reduceOnly: str = None,
    ):
        return self.place_future_market_order(
            product_symbol=product_symbol,
            side="BUY",
            quantity=quantity,
            positionSide=positionSide,
            reduceOnly=reduceOnly,
        )

    def place_future_market_sell_order(
        self,
        product_symbol: str,
        quantity: str,
        positionSide: str = None,
        reduceOnly: str = None,
    ):
        return self.place_future_market_order(
            product_symbol=product_symbol,
            side="SELL",
            quantity=quantity,
            positionSide=positionSide,
            reduceOnly=reduceOnly,
        )

    def place_future_limit_order(
        self,
        product_symbol: str,
        side: str,
        quantity: str,
        price: str,
        timeInForce: str = "GTC",
        positionSide: str = None,
        reduceOnly: str = None,
    ):
        return self.place_future_order(
            product_symbol=product_symbol,
            side=side,
            type_="LIMIT",
            quantity=quantity,
            price=price,
            timeInForce=timeInForce,
            positionSide=positionSide,
            reduceOnly=reduceOnly,
        )

    def place_future_limit_buy_order(
        self,
        product_symbol: str,
        quantity: str,
        price: str,
        timeInForce: str = "GTC",
        positionSide: str = None,
        reduceOnly: str = None,
    ):
        return self.place_future_limit_order(
            product_symbol=product_symbol,
            side="BUY",
            quantity=quantity,
            price=price,
            timeInForce=timeInForce,
            positionSide=positionSide,
            reduceOnly=reduceOnly,
        )

    def place_future_limit_sell_order(
        self,
        product_symbol: str,
        quantity: str,
        price: str,
        timeInForce: str = "GTC",
        positionSide: str = None,
        reduceOnly: str = None,
    ):
        return self.place_future_limit_order(
            product_symbol=product_symbol,
            side="SELL",
            quantity=quantity,
            price=price,
            timeInForce=timeInForce,
            positionSide=positionSide,
            reduceOnly=reduceOnly,
        )

    def place_future_post_only_limit_order(
        self,
        product_symbol: str,
        side: str,
        quantity: str,
        price: str,
        positionSide: str = None,
        reduceOnly: str = None,
    ):
        return self.place_future_order(
            product_symbol=product_symbol,
            side=side,
            type_="LIMIT",
            quantity=quantity,
            price=price,
            timeInForce="GTX",  # GTX = Post Only
            positionSide=positionSide,
            reduceOnly=reduceOnly,
        )

    def place_future_post_only_limit_buy_order(
        self,
        product_symbol: str,
        quantity: str,
        price: str,
        positionSide: str = None,
        reduceOnly: str = None,
    ):
        return self.place_future_post_only_limit_order(
            product_symbol=product_symbol,
            side="BUY",
            quantity=quantity,
            price=price,
            positionSide=positionSide,
            reduceOnly=reduceOnly,
        )

    def place_future_post_only_limit_sell_order(
        self,
        product_symbol: str,
        quantity: str,
        price: str,
        positionSide: str = None,
        reduceOnly: str = None,
    ):
        return self.place_future_post_only_limit_order(
            product_symbol=product_symbol,
            side="SELL",
            quantity=quantity,
            price=price,
            positionSide=positionSide,
            reduceOnly=reduceOnly,
        )

    def cancel_future_order(
        self,
        product_symbol: str,
        orderId: int = None,
        origClientOrderId: str = None,
    ):
        """
        :param product_symbol: str
        :param orderId: int
        :param origClientOrderId: str

        """
        payload = {
            "symbol": self.ptm.get_exchange_symbol(product_symbol, Common.BINANCE),
        }
        if orderId is not None:
            payload["orderId"] = orderId
        if origClientOrderId is not None:
            payload["origClientOrderId"] = origClientOrderId

        res = self._request(
            method="DELETE",
            path=FuturesTrade.PLACE_CANCEL_QUERY_ORDER,
            query=payload,
        )
        return res

    def get_future_order(
        self,
        product_symbol: str,
        orderId: int = None,
        origClientOrderId: str = None,
    ):
        """
        :param product_symbol: str
        :param orderId: int
        :param origClientOrderId: str

        """
        payload = {
            "symbol": self.ptm.get_exchange_symbol(product_symbol, Common.BINANCE),
        }
        if orderId is not None:
            payload["orderId"] = orderId
        if origClientOrderId is not None:
            payload["origClientOrderId"] = origClientOrderId

        res = self._request(
            method="GET",
            path=FuturesTrade.PLACE_CANCEL_QUERY_ORDER,
            query=payload,
        )
        return res

    def cancel_all_future_open_order(
        self,
        product_symbol: str,
    ):
        """
        :param product_symbol: str

        """
        payload = {
            "symbol": self.ptm.get_exchange_symbol(product_symbol, Common.BINANCE),
        }

        res = self._request(
            method="DELETE",
            path=FuturesTrade.CANCEL_ALL_OPEN_ORDERS,
            query=payload,
        )
        return res

    def get_future_all_order(
        self,
        product_symbol: str,
        orderId: int = None,
        startTime: int = None,
        endTime: int = None,
        limit: int = None,
    ):
        """
        :param product_symbol: str
        :param orderId: int
        :param startTime: int
        :param endTime: int
        :param limit: int
        """
        payload = {
            "symbol": self.ptm.get_exchange_symbol(product_symbol, Common.BINANCE),
        }
        if orderId is not None:
            payload["orderId"] = orderId
        if startTime is not None:
            payload["startTime"] = startTime
        if endTime is not None:
            payload["endTime"] = endTime
        if limit is not None:
            payload["limit"] = limit

        res = self._request(
            method="GET",
            path=FuturesTrade.QUERY_ALL_ORDERS,
            query=payload,
        )
        return res

    def get_future_open_order(
        self,
        product_symbol: str,
        orderId: int = None,
        origClientOrderId: str = None,
    ):
        """
        :param product_symbol: str

        EitherorderId or origClientOrderId must be sent
        :param orderId: int
        :param origClientOrderId: str
        """
        payload = {
            "symbol": self.ptm.get_exchange_symbol(product_symbol, Common.BINANCE),
        }
        if orderId is not None:
            payload["orderId"] = orderId
        if origClientOrderId is not None:
            payload["origClientOrderId"] = origClientOrderId

        res = self._request(
            method="GET",
            path=FuturesTrade.QUERY_OPEN_ORDER,
            query=payload,
        )
        return res

    def get_future_all_open_order(
        self,
        product_symbol: str,
    ):
        """
        :param product_symbol: str
        """
        payload = {
            "symbol": self.ptm.get_exchange_symbol(product_symbol, Common.BINANCE),
        }

        res = self._request(
            method="GET",
            path=FuturesTrade.QUERY_OPEN_ORDERS,
            query=payload,
        )
        return res

    def get_future_position(
        self,
        product_symbol: str,
    ):
        """
        :param product_symbol: str
        """
        payload = {
            "symbol": self.ptm.get_exchange_symbol(product_symbol, Common.BINANCE),
        }

        res = self._request(
            method="GET",
            path=FuturesTrade.POSITION_INFO,
            query=payload,
        )
        return res
